info = {
    'name': 'PushYourLuck',
    'description': 'As the name suggest, Push Your Luck is an artificial version of a "push your luck" game like, for instance, Can\'t Stop.',
    'context': 'ippc2018',
    'tags': '',
    'viz': 'None'
}



import numpy as np
from PIL import Image
import pygame
from pygame import gfxdraw

from pyRDDLGym.core.compiler.model import RDDLPlanningModel
from pyRDDLGym.core.visualizer.viz import BaseViz


# code comes from openai gym
class CartPoleVisualizer(BaseViz):

    def __init__(self, model: RDDLPlanningModel,
                 figure_size=(600, 400), wait_time=100) -> None:
        self._model = model
        self._figure_size = figure_size
        self._wait_time = wait_time
    
    def init_canvas(self, figure_size):
        screen = pygame.Surface(figure_size)
        surf = pygame.Surface(figure_size)
        return screen, surf
        
    def convert2img(self, screen):
        data = np.transpose(np.array(pygame.surfarray.pixels3d(screen)),
                            axes=(1, 0, 2))
        img = Image.fromarray(data)
        return img
    
    def render(self, state):
        screen, surf = self.init_canvas(self._figure_size)
        surf.fill((255, 255, 255))
        
        world_width = self._model.non_fluents['POS-LIMIT'] * 2
        scale = self._figure_size[0] / world_width
              
        # draw the cart
        cartwidth = 50.0
        cartheight = 30.0
        l, r, t, b = -cartwidth / 2, cartwidth / 2, cartheight / 2, -cartheight / 2
        axleoffset = cartheight / 4.0
        cartx = state['pos'] * scale + self._figure_size[0] / 2.0  # MIDDLE OF CART
        carty = 100  # TOP OF CART
        cart_coords = [(l, b), (l, t), (r, t), (r, b)]
        cart_coords = [(c[0] + cartx, c[1] + carty) for c in cart_coords]
        gfxdraw.aapolygon(surf, cart_coords, (0, 0, 0))
        gfxdraw.filled_polygon(surf, cart_coords, (0, 0, 0))
        
        # draw the pole
        polewidth = 10.0
        polelen = scale * (2 * self._model.non_fluents['POLE-LEN'])
        l, r, t, b = -polewidth / 2, polewidth / 2, polelen - polewidth / 2, -polewidth / 2
        pole_coords = []
        for coord in [(l, b), (l, t), (r, t), (r, b)]:
            coord = pygame.math.Vector2(coord).rotate_rad(-state['ang-pos'])
            coord = (coord[0] + cartx, coord[1] + carty + axleoffset)
            pole_coords.append(coord)
        gfxdraw.aapolygon(surf, pole_coords, (202, 152, 101))
        gfxdraw.filled_polygon(surf, pole_coords, (202, 152, 101))
        
        # draw the joints
        gfxdraw.aacircle(
            surf,
            int(cartx),
            int(carty + axleoffset),
            int(polewidth / 2),
            (129, 132, 203),
        )
        gfxdraw.filled_circle(
            surf,
            int(cartx),
            int(carty + axleoffset),
            int(polewidth / 2),
            (129, 132, 203),
        )
        gfxdraw.hline(surf, 0, self._figure_size[0], carty, (0, 0, 0))

        surf = pygame.transform.flip(surf, False, True)
        screen.blit(surf, (0, 0))        
        pygame.time.wait(self._wait_time)
        img = self.convert2img(screen)
        del screen, surf
        
        return img
    



import numpy as np
from PIL import Image
import pygame
from pygame import Rect, gfxdraw, freetype

from pyRDDLGym.core.compiler.model import RDDLPlanningModel
from pyRDDLGym.core.visualizer.viz import BaseViz

ROW_SIZE = 50
COL_SIZE = 50
ELEV_WIDTH = int(0.8 * COL_SIZE)
SIGN_RADIUS = 3

# Direction of elevators as texts
freetype.init()
DIR_TEXT = freetype.SysFont('freesansbold', 20)
NUM_IN_TEXT = freetype.SysFont('freesansbold', 20)
NUM_WAIT_TEXT = freetype.SysFont('freesansbold', 15)


# code comes from openai gym
class ElevatorVisualizer(BaseViz):

    def __init__(self, model: RDDLPlanningModel, wait_time=100) -> None:
        self._model = model
        self._wait_time = wait_time
        self._n_elev = len(self._model.type_to_objects['elevator'])
        self._n_floors = len(self._model.type_to_objects['floor'])
        self._n_cols = 7 + 2 * (self._n_elev - 1)
        self._n_rows = self._n_floors
        self._figure_size = (COL_SIZE * self._n_cols, ROW_SIZE * (self._n_rows + 2))
        self._nonfluents = model.non_fluents
        self._left = COL_SIZE 
        self._right = self._figure_size[0] - COL_SIZE
        self._top = ROW_SIZE
        self._bottom = self._figure_size[1] - ROW_SIZE
    
    def init_canvas(self, figure_size):
        screen = pygame.Surface(figure_size)
        surf = pygame.Surface(figure_size)
        return screen, surf
        
    def convert2img(self, screen):
        data = np.transpose(np.array(
            pygame.surfarray.pixels3d(screen)), axes=(1, 0, 2))
        img = Image.fromarray(data)
        return img
    
    def render(self, state):
        
        # Get the state information
        FLUENT_SEP = RDDLPlanningModel.FLUENT_SEP
        elev_to_floor = {}
        for e in self._model.type_to_objects['elevator']:
            for fl in self._model.type_to_objects['floor']:
                state_key = RDDLPlanningModel.ground_var('elevator-at-floor', [e, fl])
                if state[state_key]:
                    elev_to_floor[e] = fl
        assert len(elev_to_floor) == self._n_elev
        num_person_waiting_on_floor = 'num-person-waiting' + FLUENT_SEP + '{floor}'
        num_person_in_elevator = 'num-person-in-elevator' + FLUENT_SEP + '{elevator}'
        elev_dir_up = 'elevator-dir-up' + FLUENT_SEP + '{elevator}'
        elev_closed = 'elevator-closed' + FLUENT_SEP + '{elevator}'
        
        # Initialize the canvas
        screen, surf = self.init_canvas(self._figure_size)
        
        surf.fill((255, 255, 255))
        
        # Building exterior
        gfxdraw.box(surf, Rect(COL_SIZE // 2,
                               ROW_SIZE // 2,
                               self._figure_size[0] - COL_SIZE,
                               self._figure_size[1] - ROW_SIZE),
                               (0, 0, 0))
        gfxdraw.box(surf, Rect(self._left,
                               self._top,
                               self._right - self._left,
                               self._bottom - self._top),
                               (255, 255, 255))

        # Draw the vertical lines for elevator passages
        for i in range(self._n_elev):
            col_offset = 2 * i + 2
            gfxdraw.box(
                surf,
                Rect(self._left + col_offset * COL_SIZE,
                     self._top,
                     COL_SIZE,
                     self._bottom - self._top),
                (0, 0, 0, 50),
            )
            gfxdraw.box(
                surf,
                Rect(self._left + col_offset * COL_SIZE + int(0.08 * COL_SIZE),
                     self._top,
                     int(0.84 * COL_SIZE),
                     self._bottom - self._top),
                (255, 255, 255),
            )

        # Draw each floor
        for i in range(1, self._n_floors + 1):
            fl = self._model.type_to_objects['floor'][i - 1]
            row = i
            
            floor_y_coord = self._bottom - ROW_SIZE * row
            
            # Draw the floor level
            gfxdraw.hline(
                surf, self._left, self._right, floor_y_coord, (100, 100, 100))

            # Go through elevators
            for j in range(self._n_elev):
                elev = self._model.type_to_objects['elevator'][j]
                col_offset = 2 * j + 2
                
                if elev_to_floor[elev] == fl: 
                    n_person_elev = state[num_person_in_elevator.format(elevator=elev)]
                    e_closed = state[elev_closed.format(elevator=elev)]
                    e_up = state[elev_dir_up.format(elevator=elev)]
                    elev_left_top_coord = (
                        self._left + int((col_offset + 0.1) * COL_SIZE),
                        floor_y_coord)
                    
                    # Draw elevator
                    gfxdraw.rectangle(surf,
                                      Rect(elev_left_top_coord[0],
                                          elev_left_top_coord[1],
                                          int(COL_SIZE * 0.8),
                                          ROW_SIZE),
                                      (0, 0, 0))
                    if e_closed:
                        gfxdraw.box(surf,
                                    Rect(elev_left_top_coord[0],
                                        elev_left_top_coord[1],
                                        int(COL_SIZE * 0.8),
                                        ROW_SIZE),
                                    (0, 0, 0, 100))
                        gfxdraw.vline(
                            surf,
                            elev_left_top_coord[0] + ELEV_WIDTH // 2,
                            floor_y_coord,
                            floor_y_coord + ROW_SIZE,
                            (0, 0, 0)
                        )
                    else:
                        gfxdraw.box(surf,
                                    Rect(elev_left_top_coord[0],
                                        elev_left_top_coord[1],
                                        int(COL_SIZE * 0.8),
                                        ROW_SIZE),
                                    (0, 0, 0, 100))
                        gfxdraw.box(surf,
                                    Rect(elev_left_top_coord[0] + int(0.1 * COL_SIZE),
                                        elev_left_top_coord[1] + int(0.1 * ROW_SIZE),
                                        int(COL_SIZE * 0.6),
                                        int(ROW_SIZE * 0.8)),
                                    (255, 255, 255))
                        
                    # Display the direction
                    DIR_TEXT.render_to(surf,
                                       (int(elev_left_top_coord[0] + 0.15 * COL_SIZE),
                                        int(elev_left_top_coord[1] + 0.1 * ROW_SIZE)),
                                       "^" if e_up else "v",
                                       (0, 0, 0))
                    NUM_IN_TEXT.render_to(surf,
                                       (int(elev_left_top_coord[0] + 0.15 * COL_SIZE),
                                        int(elev_left_top_coord[1] + 0.4 * ROW_SIZE)),
                                       str(n_person_elev),
                                       (255, 0, 0) if n_person_elev > 0 else (0, 0, 255)
                                       )
                    gfxdraw.aacircle(
                        surf,
                        (elev_left_top_coord[0] + ELEV_WIDTH // 2),
                        elev_left_top_coord[1] - SIGN_RADIUS,
                        SIGN_RADIUS,
                        (255, 0, 0) if e_closed else (0, 255, 0),
                    )
                    gfxdraw.filled_circle(
                        surf,
                        (elev_left_top_coord[0] + ELEV_WIDTH // 2),
                        elev_left_top_coord[1] - SIGN_RADIUS,
                        SIGN_RADIUS,
                        (255, 0, 0) if e_closed else (0, 255, 0),
                    )

            n_person_waiting = state[num_person_waiting_on_floor.format(floor=fl)]
            col_offset = self._n_cols - 3.1
            NUM_IN_TEXT.render_to(
                surf,
                (self._left + int((col_offset) * COL_SIZE),
                 floor_y_coord + 0.5 * ROW_SIZE),
                str(n_person_waiting),
                (255, 0, 0) if n_person_waiting > 0 else (0, 0, 255)
            )
            
        # surf = pygame.transform.flip(surf, False, True)
        screen.blit(surf, (0, 0))        
        pygame.time.wait(self._wait_time)        
        img = self.convert2img(screen)        
        del screen, surf        
        
        return img
    



import io
import matplotlib.pyplot as plt
import numpy as np
from PIL import Image
import pygame

from pyRDDLGym.core.compiler.model import RDDLPlanningModel
from pyRDDLGym.core.visualizer.viz import BaseViz


class QuadcopterVisualizer(BaseViz):

    def __init__(self, model: RDDLPlanningModel,
                 bound=4,
                 figure_size=(6, 6),
                 wait_time=10) -> None:
        self._model = model
        self._bound = bound
        self._figure_size = figure_size
        self._wait_time = wait_time
        
        self._objects = model.type_to_objects
        self._nonfluents = model.ground_vars_with_values(model.non_fluents)
        
        self.fig = plt.figure(figsize=self._figure_size)
        self.ax = self.fig.add_subplot(111, projection='3d')
        
        self.xs, self.ys, self.zs = {}, {}, {}
        for dk in self._objects['drone']:
            self.xs[dk] = []
            self.ys[dk] = []
            self.zs[dk] = []
            
    def convert2img(self):
        img_buf = io.BytesIO()
        self.fig.savefig(img_buf, format='png')
        im = Image.open(img_buf)
        return im
    
    def render(self, state):      
        
        # draw the drones  
        for dk in self._objects['drone']:
            
            # convert state to position information
            pos = np.array([state[f'x___{dk}'], state[f'y___{dk}'], state[f'z___{dk}']])
            phi, theta, psi = state[f'phi___{dk}'], state[f'theta___{dk}'], state[f'psi___{dk}']
            cph, cth, cps = np.cos(phi), np.cos(theta), np.cos(psi)
            sph, sth, sps = np.sin(phi), np.sin(theta), np.sin(psi)        
            R = np.array([
                [cph * cps - cth * sth * sps, -cps * sph - cph * cth * sps, sth * sps],
                [cth * cps * sph + cph * sps, cph * cth * cps - sph * sps, -cps * sth],
                [sph * sth, cph * sth, cth]
            ])
            l = self._nonfluents['L']
            motor1 = (R @ np.array([[l], [0], [0]])).reshape((-1,)) + pos
            motor2 = (R @ np.array([[-l], [0], [0]])).reshape((-1,)) + pos
            motor3 = (R @ np.array([[0], [l], [0]])).reshape((-1,)) + pos
            motor4 = (R @ np.array([[0], [-l], [0]])).reshape((-1,)) + pos
            perp = (R @ np.array([[0], [0], [2 * l]])).reshape((-1,)) + pos
            
            tpos = np.array([self._nonfluents[f'TX___{dk}'],
                             self._nonfluents[f'TY___{dk}'],
                             self._nonfluents[f'TZ___{dk}']])
            
            # draw the drone
            self.ax.scatter(*pos, c='black', s=(40 * self._nonfluents['R']) ** 2)
            self.ax.scatter(*motor1, c='black', s=(20 * self._nonfluents['R']) ** 2)
            self.ax.scatter(*motor2, c='black', s=(20 * self._nonfluents['R']) ** 2)
            self.ax.scatter(*motor3, c='black', s=(20 * self._nonfluents['R']) ** 2)
            self.ax.scatter(*motor4, c='black', s=(20 * self._nonfluents['R']) ** 2)
            self.ax.plot(*zip(pos, motor1), c='black')
            self.ax.plot(*zip(pos, motor2), c='black')
            self.ax.plot(*zip(pos, motor3), c='black')
            self.ax.plot(*zip(pos, motor4), c='black')
            self.ax.plot(*zip(pos, perp), c='red', alpha=0.4)
            self.ax.scatter(*tpos, c='red', marker='>', s=40)
            
            self.xs[dk].append(pos[0])
            self.ys[dk].append(pos[1])
            self.zs[dk].append(pos[2])
            
            self.ax.plot(self.xs[dk], self.ys[dk], self.zs[dk], 
                         color='blue', alpha=0.1, linewidth=1.0)
        
        self.ax.set_xlim3d([-self._bound, self._bound])
        self.ax.set_ylim3d([-self._bound, self._bound])
        self.ax.set_zlim3d([-self._bound, self._bound])
        self.ax.xaxis.set_pane_color((0.0, 0.0, 1.0, 0.0))
        self.ax.yaxis.set_pane_color((0.0, 0.0, 1.0, 0.0))
        self.ax.zaxis.set_pane_color((0.0, 0.0, 1.0, 0.0))
        self.ax.tick_params(axis='x', colors='dimgray')
        self.ax.tick_params(axis='y', colors='dimgray')
        self.ax.tick_params(axis='z', colors='dimgray')
        self.ax.xaxis._axinfo['grid']['linewidth'] = 0.1
        self.ax.yaxis._axinfo['grid']['linewidth'] = 0.1
        self.ax.zaxis._axinfo['grid']['linewidth'] = 0.1

        plt.rcParams.update({'font.size': 12, 'font.family': 'Consolas'})
        plt.tight_layout()
        
        pygame.time.wait(self._wait_time)        
        img = self.convert2img()        
        self.ax.cla()
        
        return img




info = {
    'name': 'TrafficBLX_SimplePhases',
    'description': 'BLX/QTM traffic signal control model with a fixed phase progression consisting of 4 phases. The goal is to control traffic lights to minimize total travel time.',
    'context': '',
    'tags': '',
    'viz': 'TrafficViz.TrafficVisualizer'
}



import matplotlib.pyplot as plt
from matplotlib.patches import Arrow, Rectangle
from matplotlib import collections  as mc
import numpy as np
from PIL import Image

from pyRDDLGym.core.compiler.model import RDDLPlanningModel
from pyRDDLGym.core.visualizer.viz import BaseViz


class IntrudersVisualizer(BaseViz):

    def __init__(self, model: RDDLPlanningModel,
                 figure_size=(4, 4),
                 vector_len=0.15,
                 lb=-2, ub=2,
                 wait_time=100) -> None:
        self._model = model
        self._figure_size = figure_size
        self._vector_len = vector_len
        self._wait_time = wait_time
        
        self._nonfluents = model.ground_vars_with_values(model.non_fluents)
        
        self.fig = plt.figure(figsize=self._figure_size)
        self.ax = plt.gca()
        
        # draw zones
        FLUENT_SEP = RDDLPlanningModel.FLUENT_SEP
        obj = model.type_to_objects['zone']
        lines = []
        for o in obj:
            l = self._nonfluents['DANGER-L' + FLUENT_SEP + o]
            r = self._nonfluents['DANGER-R' + FLUENT_SEP + o]
            t = self._nonfluents['DANGER-T' + FLUENT_SEP + o]
            b = self._nonfluents['DANGER-B' + FLUENT_SEP + o]
            zone = Rectangle((l, b), r - l, t - b, color='orange', alpha=0.4)
            self.ax.add_patch(zone)
            
        #     lines.append([(l, t), (r, t)])
        #     lines.append([(l, t), (l, b)])
        #     lines.append([(r, t), (r, b)])
        #     lines.append([(l, b), (r, b)])
        # lc = mc.LineCollection(lines, linewidths=2)
        # self.ax.add_collection(lc)
        
    def convert2img(self, canvas):
        data = np.frombuffer(canvas.buffer_rgba(), dtype=np.uint8)
        data = data.reshape(canvas.get_width_height()[::-1] + (4,))
        data = data[:, :, :3]
        return Image.fromarray(data)

    def render(self, state):
        FLUENT_SEP = RDDLPlanningModel.FLUENT_SEP
        
        camera_circs, camera_arrows = [], []
        for camera in self._model.type_to_objects['camera']: 
            x = state['camera-x' + FLUENT_SEP + camera]
            y = state['camera-y' + FLUENT_SEP + camera]
            vx = state['camera-vx' + FLUENT_SEP + camera]
            vy = state['camera-vy' + FLUENT_SEP + camera]            
            circ = plt.Circle((x, y), 
                              self._nonfluents['CAMERA-RADIUS' + FLUENT_SEP + camera],
                              color='green', alpha=0.4)
            camera_circs.append(self.ax.add_patch(circ))
            arrow = Arrow(x, y, vx, vy, width=0.1, color='green')
            camera_arrows.append(self.ax.add_patch(arrow))
            
        intr_circs, intr_arrows = [], []
        for intr in self._model.type_to_objects['intruder']: 
            x = state['intruder-x' + FLUENT_SEP + intr]
            y = state['intruder-y' + FLUENT_SEP + intr]
            vx = state['intruder-vx' + FLUENT_SEP + intr]
            vy = state['intruder-vy' + FLUENT_SEP + intr]            
            circ = plt.Circle((x, y), 0.01, color='red')
            intr_circs.append(self.ax.add_patch(circ))
            arrow = Arrow(x, y, vx, vy, width=0.05, color='red')
            intr_arrows.append(self.ax.add_patch(arrow))
                
        self.fig.canvas.draw()
        img = self.convert2img(self.fig.canvas)   
        
        for obj in camera_circs + camera_arrows + intr_circs + intr_arrows:
            obj.remove()
        del camera_circs, camera_arrows, intr_circs, intr_arrows
        
        return img



info = {
    'name': 'TowerOfHanoi',
    'description': 'The classic tower of Hanoi puzzle, where disks must be stacked onto a given rod.',
    'context': 'arcade',
    'tags': '',
    'viz': 'None'
}


import matplotlib
matplotlib.use('agg')
import pygame
import matplotlib.pyplot as plt
from matplotlib.path import Path
import matplotlib.patches as mpatches
import numpy as np
from PIL import Image
from collections import defaultdict
from copy import copy
import math

from pyRDDLGym.core.compiler.model import RDDLPlanningModel
from pyRDDLGym.core.visualizer.viz import BaseViz

ROAD_PAVED_WIDTH = 12  # The width of the paved part of the road in m
ROAD_HALF_MEDIAN_WIDTH = 1  # The width of the median in-between links in m
ROAD_WIDTH = ROAD_PAVED_WIDTH + ROAD_HALF_MEDIAN_WIDTH  # Combined width of a single link (paved + 0.5*median)
ROAD_DELTA = ROAD_PAVED_WIDTH / 12  # Division of a link into parts for drawing turns
TURN_RADIUS_DELTA = 5  # Curvature of a turn
LINK_GRID_SAT_DENSITY = 0.2  # per lane                   # Density at which a cell in the flow grid gets a saturated color

RGBA_PAVEMENT_COLOR = (0.6, 0.6, 0.6, 1)
RGBA_INTERSECTION_COLOR = (0.3, 0.3, 0.3, 1)
RGBA_GREEN_TURN_COLOR = (0.2, 1, 0.2, 1)
RGBA_Q_PATCH_COLOR = (1, 0.2, 0.2, 0.4)
RGBA_Q_LINE_COLOR = (1, 0.2, 0.2, 1)
RGBA_LINK_GRID_COLOR = RGBA_PAVEMENT_COLOR
RGBA_LINK_GRID_SAT_COLOR = (0.9, 0.3, 0.9, 1)


def get_fillcolor_by_density(N, capacity):
    p = max(min(1, N / capacity), 0)
    return tuple(p * c for c in RGBA_LINK_GRID_SAT_COLOR)


def id(var_id, *objs):
    # Returns <var_id>___<ob1>__<ob2>__<ob3>...
    return var_id + RDDLPlanningModel.FLUENT_SEP + RDDLPlanningModel.OBJECT_SEP.join(objs)


class TrafficVisualizer(BaseViz):

    def __init__(self,
                 model: RDDLPlanningModel,
                 figure_size=None,  # Pass [int, int] when overriding default
                 dpi=100,
                 fontsize=10,
                 display=False) -> None:
        self._model = model
        self._states = model.ground_vars_with_values(model.state_fluents)
        self._nonfluents = model.ground_vars_with_values(model.non_fluents)
        self._objects = model.type_to_objects

        if figure_size is None:
            pygame.init()
            info = pygame.display.Info()
            if info.current_w == -1 or info.current_h == -1:
                # Display info not available to pygame
                self._figure_size = [10, 10]
            else:
                self._figure_size = [0.9 * info.current_w / dpi, 0.9 * info.current_h / dpi]

        self._display = display
        self._dpi = dpi
        self._fig, self._ax = None, None
        self._fontsize = fontsize
        self._interval = 10
        self._nonfluents_layout = None
        self._states_layout = None
        self._data = None
        self._img = None

        self.parse_nonfluents()
        self.build_nonfluent_patches()

    def parse_nonfluents(self):
        self.veh_len = self._nonfluents['Lv']
        self.sim_step = self._nonfluents['Ts']

        # Register the intersection coordinates and the links incident to them
        self.intersections = {}
        self.linkdata = defaultdict(dict)
        for d in self._objects['intersection']:
            coords = np.array([self._nonfluents[id('X', d)],
                               self._nonfluents[id('Y', d)]])
            inc_links, out_links = [], []

            for L in self._objects['link']:

                if self._nonfluents[id('LINK-TO', L, d)]:
                    self.linkdata[L]['to'] = coords
                    self.linkdata[L]['to_id'] = d
                    inc_links.append(L)

                elif self._nonfluents[id('LINK-FROM', d, L)]:
                    self.linkdata[L]['from'] = coords
                    self.linkdata[L]['from_id'] = d
                    out_links.append(L)

            self.intersections[d] = {
                'coords': coords,
                'inc_links': inc_links,
                'out_links': out_links}

        # Register the links and record the required link data
        ccw_turn_matrix = np.array([[0, -1], [1, 0]])
        self.sources, self.sinks = {}, {}

        for L in self.linkdata:

            # Register the sources and sinks, and in addition
            # find the start and ends points of the link
            if self._nonfluents[id('SOURCE', L)]:
                try:
                    vu = np.array([self._nonfluents[id('SOURCE-X', L)],
                                   self._nonfluents[id('SOURCE-Y', L)]])
                except KeyError as e:
                    print('[TrafficViz] Please include the X and Y coordinates of the source '
                          f'{L} in the instance defn')
                    raise e
                self.sources[L] = {'coords': vu}
                self.linkdata[L]['from'] = vu
                self.linkdata[L]['from_id'] = f'source{len(self.sources)}'

            elif self._nonfluents[id('SINK', L)]:
                try:
                    vd = np.array([self._nonfluents[id('SINK-X', L)],
                                   self._nonfluents[id('SINK-Y', L)]])
                except KeyError as e:
                    print('[TrafficViz] Please include the X and Y coordinates of the sink '
                          f'{L} in the instance defn')
                    raise e
                self.sinks[L] = {'coords': vd}
                self.linkdata[L]['to'] = vd
                self.linkdata[L]['to_id'] = f'sink{len(self.sinks)}'

            vu, vd = self.linkdata[L]['from'], self.linkdata[L]['to']
            dir = vd - vu
            link_len = np.linalg.norm(dir)

            num_lanes = self._nonfluents[id('Nl', L)]
            v = self._nonfluents[id('Vl', L)]
            num_cells = math.ceil(link_len / (v * self.sim_step))
            cell_capacity = (link_len * num_lanes) / (num_cells * self.veh_len)

            dir = dir / link_len

            self.linkdata[L].update({
                # Unit vector along the link direction
                'dir': dir,
                # Unit normal vector (ccw of unit dir vector)
                'nrm': np.matmul(ccw_turn_matrix, dir),
                # Length of the link
                'len': link_len,
                # Freeflow speed on the link
                'v': v,
                # Number of lanes
                'num_lanes': num_lanes,
                # Number of incoming cells
                'num_cells': num_cells,
                # Cell capacity (for shading)
                'cell_capacity': cell_capacity })

        # Register the turning movements.
        # Turns are indexed by upstream, downstream link pairs
        # Additionally, order turns from each link left-to-right
        def signed_angle(v0, v1):
            if np.cross(v0, v1) == 0:
                return -math.pi if np.dot(v0, v1) < 0 else 0
            return np.sign(np.cross(v0, v1)) * np.arccos(np.dot(v0, v1))

        self.turn_ids = []
        for L in self.linkdata:
            self.linkdata[L]['turns_from'] = []
            for M in self.linkdata:
                if self._nonfluents[id('TURN', L, M)]:
                    self.turn_ids.append((L, M))
                    self.linkdata[L]['turns_from'].append(M)
            self.linkdata[L]['turns_from'].sort(
                key=lambda M:-signed_angle(self.linkdata[L]['dir'], self.linkdata[M]['dir']))

        # Try to find the opposite link by the most negative dot product
        for d, v in self.intersections.items():
            for L in v['inc_links']:
                dL = self.linkdata[L]['dir']
                M = sorted(v['out_links'], key=lambda M: np.dot(dL, self.linkdata[M]['dir']))[0]
                self.linkdata[L]['opposite_link'] = M
                self.linkdata[M]['opposite_link'] = L

        # Register the signal phases
        self.signal_phases = self._objects['signal-phase']

        # Register the green turns for each phase in each intersection
        self.green_turns_by_intersection_phase = {}
        for d, dv in self.intersections.items():
            self.green_turns_by_intersection_phase[d] = {k: [] for k in self.signal_phases}
            for L in dv['inc_links']:
                for M in self.linkdata[L]['turns_from']:
                    for p in self.signal_phases:
                        if self._nonfluents[id('GREEN', L, M, p)]:
                            self.green_turns_by_intersection_phase[d][p].append((L, M))

    def build_nonfluent_patches(self):
        # Find the bbox
        x0, y0, x1, y1 = math.inf, math.inf, -math.inf, -math.inf
        for geopoint in (tuple(self.intersections.values()) +
                         tuple(self.sources.values()) +
                         tuple(self.sinks.values())):
            v = geopoint['coords']
            x0, y0, x1, y1 = min(x0, v[0]), min(y0, v[1]), max(x1, v[0]), max(y1, v[1])

        # Slightly expand bbox
        w, h = x1 - x0, y1 - y0
        dw, dh = 0.05 * w, 0.05 * h
        self.bbox = {'x0': x0 - dw, 'x1': x1 + dw,
                     'y0': y0 - dh, 'y1': y1 + dh,
                     'w': round(w + 2 * dw), 'h': round(h + 2 * dh)}

        def quad_bezier_middle_ctrl_pt(ctrl0, ctrl2, dir0, dir2):
            # Find the middle control point of a quadratic Bezier curve,
            # given the first and third control points ctrl0 and ctrl2,
            # and the tangent vectors d0 at ctrl0 and d2 at ctrl2
            c = np.cross(dir0, dir2)
            if c == 0:
                # Exceptional case when the tangent vectors are colinear
                ctrl1 = (ctrl0 + ctrl2) / 2
            else:
                ctrl1 = ctrl2 + (np.cross(dir0, ctrl0 - ctrl2) / c) * dir2
            return ctrl1

        # Construct intersection objects and correct link lengths
        for d in self.intersections:

            # Preliminary computations
            alpha_in, alpha_out = {}, {}
            for L in self.intersections[d]['inc_links']:
                # Right-turn-target from L
                M = self.linkdata[L]['turns_from'][-1]

                # Find the point of intersection of the right boundary of L
                # and the right boundary of M (if both links are extended).
                # The links will be shortened so that they do not intersect,
                # as well as shortened by TURN_RADIUS_DELTA to draw a smoother
                # connecting curve
                vu, vd = self.linkdata[L]['from'], self.linkdata[L]['to']
                dir0, n0 = self.linkdata[L]['dir'], self.linkdata[L]['nrm']
                dir1, n1 = self.linkdata[M]['dir'], self.linkdata[M]['nrm']
                c = np.cross(dir0, dir1)
                if c == 0:
                    # Exceptional case when the right-turn direction happens to be straight ahead
                    alpha_in[L] = 0
                    alpha_out[M] = 0
                else:
                    alpha_in[L] = ROAD_WIDTH * np.cross(dir1, n1 - n0) / np.cross(dir0, dir1)
                    alpha_out[M] = ROAD_WIDTH * np.cross(dir0, n1 - n0) / np.cross(dir0, dir1)

            # Construct the intersection contour
            L0 = self.intersections[d]['inc_links'][0]
            L = L0
            contour, codes = [], []
            while True:

                # Right-turn-target from L
                M = self.linkdata[L]['turns_from'][-1]
                Lopp = self.linkdata[L]['opposite_link']
                Mopp = self.linkdata[M]['opposite_link']

                vu, vd = self.linkdata[L]['from'], self.linkdata[L]['to']
                dir0, n0 = self.linkdata[L]['dir'], self.linkdata[L]['nrm']
                dir1, n1 = self.linkdata[M]['dir'], self.linkdata[M]['nrm']

                # Find the control points for the right-turn Bezier curves
                # Shorten the links by the same amount in both travel directions
                beta = max(-alpha_in[L], alpha_out[Lopp]) + TURN_RADIUS_DELTA  # L shortening
                alpha = max(-alpha_in[Mopp], alpha_out[M]) + TURN_RADIUS_DELTA  # M shortening

                # Right-turn Bezier control points
                ctrl0 = vd - ROAD_WIDTH * n0 - beta * dir0
                ctrl2 = vd - ROAD_WIDTH * n1 + alpha * dir1
                ctrl1 = quad_bezier_middle_ctrl_pt(ctrl0, ctrl2, dir0, dir1)

                # Correct the link lengths
                # * Subtract TURN_RADIUS_DELTA
                # * Avoid intersections
                self.linkdata[L]['len'] -= beta
                self.linkdata[M]['from'] = vd + alpha * dir1
                self.linkdata[M]['len'] -= alpha

                # Trace the path
                F = self.linkdata[L].get('from', vu)
                Llen = self.linkdata[L]['len']
                p = F + Llen * dir0 - ROAD_HALF_MEDIAN_WIDTH * n0
                if len(contour) == 0:
                    # Initializer
                    contour.append(p)
                    codes.append(Path.MOVETO)
                contour.append(p)
                codes.append(Path.LINETO)
                contour.extend([ctrl0, ctrl1, ctrl2])
                codes.extend([Path.LINETO, Path.CURVE3, Path.CURVE3])
                p = ctrl2 + ROAD_PAVED_WIDTH * n1
                contour.append(p)
                codes.append(Path.LINETO)
                ctrl0 = p
                ctrl2 = p + 2 * ROAD_HALF_MEDIAN_WIDTH * n1
                ctrl1 = p + ROAD_HALF_MEDIAN_WIDTH * n1 - TURN_RADIUS_DELTA * dir1
                contour.extend([ctrl0, ctrl1, ctrl2])
                codes.extend([Path.LINETO, Path.CURVE3, Path.CURVE3])

                L = Mopp
                if L == L0:
                    contour.append(np.array([0, 0]))  # ignored
                    codes.append(Path.CLOSEPOLY)
                    break

            # Done with the intersection contour!
            path = Path(contour, codes)
            patch = mpatches.PathPatch(
                path,
                color=RGBA_INTERSECTION_COLOR)
            self.intersections[d]['patch'] = patch

        # Construct the link patches,
        # register data for plotting the queues,
        # construct turning curve patches
        self.link_patches = {}
        self.turn_patches = {}
        for L, Lv in self.linkdata.items():
            vu, vd = Lv['from'], Lv['to']
            d0, n0 = Lv['dir'], Lv['nrm']
            road = np.zeros(shape=(4, 2))
            Llen = Lv['len']

            # Record the visual shrinking factor for visualizing queues
            self.linkdata[L]['shrink'] = Llen / np.linalg.norm(vd - vu)

            # Create the link patch
            road[0] = vu - ROAD_HALF_MEDIAN_WIDTH * n0
            road[1] = road[0] + Llen * d0
            road[2] = road[1] - ROAD_PAVED_WIDTH * n0
            road[3] = road[2] - Llen * d0
            self.link_patches[L] = plt.Polygon(road, color=RGBA_PAVEMENT_COLOR)

            self.linkdata[L]['cell_len'] = Llen / self.linkdata[L]['num_cells']
            self.linkdata[L]['stopline_left'] = road[1]

            for idx, M in enumerate(Lv['turns_from']):
                # Turns are indexed left-to-right
                d1, n1 = self.linkdata[M]['dir'], self.linkdata[M]['nrm']

                # Turning curve patch
                delta = (2 + 4 * idx) * ROAD_DELTA
                from_ = road[1] - delta * n0
                to_ = self.linkdata[M]['from'] - (ROAD_HALF_MEDIAN_WIDTH + delta) * n1
                ctrl1 = quad_bezier_middle_ctrl_pt(from_, to_, d0, d1)
                path = Path([from_, ctrl1, to_], [Path.MOVETO, Path.CURVE3, Path.CURVE3])
                patch = mpatches.PathPatch(
                    path,
                    edgecolor=RGBA_GREEN_TURN_COLOR,
                    fill=False,
                    linewidth=2)
                self.turn_patches[(L, M)] = patch

    def init_canvas(self, figure_size, dpi):
        fig = plt.figure(figsize=figure_size, dpi=dpi)
        ax = plt.gca()
        plt.xlim(self.bbox['x0'], self.bbox['x1'])
        plt.ylim(self.bbox['y0'], self.bbox['y1'])
        plt.axis('scaled')
        plt.axis('off')
        return fig, ax

    def convert2img(self, fig, ax):
        ax.set_position((0, 0, 1, 1))
        fig.canvas.draw()
        data = np.frombuffer(fig.canvas.buffer_rgba(), dtype=np.uint8)
        data = data.reshape(fig.canvas.get_width_height()[::-1] + (4,))
        data = data[:, :, :3]
        img = Image.fromarray(data)
        self._data = data
        self._img = img
        return img

    def build_nonfluents_layout(self, fig, ax):
        # Seems copy is the only way to re-use patches
        # when using more than one axes object :(
        for link_patch in self.link_patches.values():
            ax.add_patch(copy(link_patch))

        for v in self.intersections.values():
            ax.add_patch(copy(v['patch']))

    def build_states_layout(self, states, fig, ax):

        for d in self.intersections:
            if states[id('all-red', d)] > 0.0:
                continue
            signal = self.signal_phases[states[id('signal', d)]]
            for t in self.green_turns_by_intersection_phase[d][signal]:
                ax.add_patch(copy(self.turn_patches[t]))

        for L, Lv in self.linkdata.items():
            # If the link is not a sink, draw the total queue
            # as well as the queue proportions by turning movement.
            # (for a sink, there is no queue)
            Q = sum(states[id('q', L, M)] for M in Lv['turns_from'])

            if L not in self.sinks and Q > 0:
                Q_dist = (Q * self.veh_len / Lv['num_lanes']) * Lv['shrink']
                Q_frontier_L = Lv['stopline_left'] - Q_dist * Lv['dir']
                Q_frontier_R = Q_frontier_L - ROAD_PAVED_WIDTH * Lv['nrm']

                # Draw the total queue patch
                Q_total_patch = np.zeros(shape=(4, 2))
                Q_total_patch[0] = Lv['stopline_left']
                Q_total_patch[1] = Q_total_patch[0] - ROAD_PAVED_WIDTH * Lv['nrm']
                Q_total_patch[2] = Q_frontier_R
                Q_total_patch[3] = Q_frontier_L
                patch = plt.Polygon(Q_total_patch, facecolor=RGBA_Q_PATCH_COLOR, linewidth=1)
                ax.add_patch(patch)

                # Draw the queue-proportions per individual turn
                for idx, M in enumerate(Lv['turns_from']):
                    Q_turn = states[id('q', L, M)]
                    barlen = (Q_turn / Q) * Q_dist
                    Q_turn_bar = np.zeros(shape=(4, 2))
                    Q_turn_bar[0] = Lv['stopline_left'] - (1 + 4 * idx) * ROAD_DELTA * Lv['nrm']
                    Q_turn_bar[1] = Q_turn_bar[0] - 2 * ROAD_DELTA * Lv['nrm']
                    Q_turn_bar[2] = Q_turn_bar[1] - barlen * Lv['dir']
                    Q_turn_bar[3] = Q_turn_bar[2] + 2 * ROAD_DELTA * Lv['nrm']
                    patch = plt.Polygon(Q_turn_bar, color=RGBA_Q_LINE_COLOR, linewidth=0)
                    ax.add_patch(patch)
            else:
                Q_dist, Q_frontier_L, Q_frontier_R = 0, Lv['stopline_left'], Lv['stopline_left'] - ROAD_PAVED_WIDTH * Lv['nrm']

            # Plot the flows through the link
            num_cells_in_use = math.ceil((Lv['len'] - Q_dist) / Lv['cell_len'])
            frontier = np.array([Q_frontier_L, Q_frontier_R])
            saturation_d = LINK_GRID_SAT_DENSITY * Lv['num_lanes']
            for t in range(1, num_cells_in_use):
                flow = states[id('flow-on-link', L, f't{t}')]
                color = get_fillcolor_by_density(flow, saturation_d)
                next_frontier = frontier - Lv['cell_len'] * Lv['dir']
                cell = np.array((frontier[0], frontier[1], next_frontier[1], next_frontier[0]))
                patch = plt.Polygon(cell, linewidth=1, facecolor=color,
                                    edgecolor=RGBA_LINK_GRID_COLOR)
                ax.add_patch(patch)
                frontier = next_frontier
            # The last cell may be shorter
            flow = states[id('flow-on-link', L, f't{num_cells_in_use}')]
            color = get_fillcolor_by_density(flow, saturation_d)
            delta = Lv['len'] - (Lv['cell_len'] * (num_cells_in_use - 1) + Q_dist)
            next_frontier = frontier - delta * Lv['dir']
            cell = np.array((frontier[0], frontier[1], next_frontier[1], next_frontier[0]))
            patch = plt.Polygon(cell, linewidth=1, facecolor=color,
                                edgecolor=RGBA_LINK_GRID_COLOR)
            ax.add_patch(patch)

    def render(self, state):
        self.states = state

        self._fig, self._ax = self.init_canvas(self._figure_size, self._dpi)
        self.build_nonfluents_layout(self._fig, self._ax)
        self.build_states_layout(state, self._fig, self._ax)

        img = self.convert2img(self._fig, self._ax)
        self._ax.cla()
        plt.close()
        return img



import matplotlib.pyplot as plt
from matplotlib.patches import Arrow
from matplotlib import collections  as mc
import numpy as np
from PIL import Image

from pyRDDLGym.core.compiler.model import RDDLPlanningModel
from pyRDDLGym.core.visualizer.viz import BaseViz


class RacecarVisualizer(BaseViz):

    def __init__(self, model: RDDLPlanningModel,
                 figure_size=(4, 4),
                 car_radius=0.04,
                 vector_len=0.15,
                 wait_time=100) -> None:
        self._model = model
        self._figure_size = figure_size
        self._car_radius = car_radius
        self._vector_len = vector_len
        self._wait_time = wait_time
        
        self._nonfluents = model.ground_vars_with_values(model.non_fluents)
        
        self.fig = plt.figure(figsize=self._figure_size)
        self.ax = plt.gca()
        
        # draw boundaries
        FLUENT_SEP = RDDLPlanningModel.FLUENT_SEP
        obj = model.type_to_objects['b']
        X1, X2, Y1, Y2 = [], [], [], []
        for o in obj:
            X1.append(self._nonfluents['X1' + FLUENT_SEP + o])
            X2.append(self._nonfluents['X2' + FLUENT_SEP + o])
            Y1.append(self._nonfluents['Y1' + FLUENT_SEP + o])
            Y2.append(self._nonfluents['Y2' + FLUENT_SEP + o])
        lines = [[(x1, y1), (x2, y2)] for x1, y1, x2, y2 in zip(X1, Y1, X2, Y2)]
        lc = mc.LineCollection(lines, linewidths=2)
        self.ax.add_collection(lc)
        self.ax.autoscale()
        self.ax.margins(0.1)
        
        # draw goal        
        goal = plt.Circle((self._nonfluents['GX'], self._nonfluents['GY']),
                          self._nonfluents['RADIUS'],
                          color='g')
        self.ax.add_patch(goal)
        
        # velocity vector
        self.arrow = Arrow(0, 0, 0, 0, 
                           width=0.2 * self._vector_len, 
                           color='black')
        self.move = self.ax.add_patch(self.arrow)

    def convert2img(self, canvas):
        data = np.frombuffer(canvas.buffer_rgba(), dtype=np.uint8)
        data = data.reshape(canvas.get_width_height()[::-1] + (4,))
        data = data[:, :, :3]
        return Image.fromarray(data)

    def render(self, state):
        vel = np.array([state['vx'], state['vy']])
        if np.max(np.abs(vel)) > 0:
            vel /= np.linalg.norm(vel)
        vel *= self._vector_len
        dx = vel[0]
        dy = vel[1]
        self.move.remove()
        self.arrow = Arrow(state['x'], state['y'], dx, dy, 
                           width=0.2 * self._vector_len, 
                           color='black')
        self.move = self.ax.add_patch(self.arrow)

        car = plt.Circle((state['x'], state['y']), self._car_radius)
        self.ax.add_patch(car)
        
        self.fig.canvas.draw()
        img = self.convert2img(self.fig.canvas)        
        car.remove()
        del car
        
        return img




info = {
    'name': 'MountainCar',
    'description': 'A simple continuous MDP for the classical mountain car control problem',
    'context': 'ippc2023',
    'tags': '',
    'viz': 'MountainCarViz.MountainCarVisualizer',
    'generator': 'MountainCarInstanceGenerator.MountainCarInstanceGenerator'
}


info = {
    'name': 'Reservoir',
    'description': 'Continuous action version of management of the water level in interconnected reservoirs',
    'context': 'ippc2023',
    'tags': '',
    'viz': 'ReservoirViz.ReservoirVisualizer',
    'generator': 'ReservoirInstanceGenerator.ReservoirInstanceGenerator'
}


info = {
    'name': 'PowerGen',
    'description': 'A continuous simple power generation problem loosely modeled on the problem of unit commitment',
    'context': 'ippc2023',
    'tags': '',
    'viz': 'PowerGenViz.PowerGenVisualizer',
    'generator': 'PowerGenInstanceGenerator.PowerGenInstanceGenerator'
}


"""A tool for generating instances of the Traffic domain"""

import numpy as np
from itertools import product

indent_str = ' ' * 8
newline_indent_str = '\n' + indent_str

def dist(p0, p1):
    return np.linalg.norm(p1-p0)


def generate_simple_4leg_intersection(i, Ein, Nout, Nin, Wout, Win, Sout, Sin, Eout,
                                      min_left, min_through, max, red, right_on_red=True):
    """Generates the non-fluents for a four-phase four-leg intersection,
    with left, through, and right movements"""

    nonfluents_str = newline_indent_str*2 + f'//intersection {i}'
    nonfluents_str += newline_indent_str + '//turns' + newline_indent_str
    nonfluents_str += newline_indent_str.join((
        f'TURN({Ein},{Nout});',
        f'TURN({Ein},{Wout});',
        f'TURN({Ein},{Sout});',
        f'TURN({Nin},{Wout});',
        f'TURN({Nin},{Sout});',
        f'TURN({Nin},{Eout});',
        f'TURN({Win},{Sout});',
        f'TURN({Win},{Eout});',
        f'TURN({Win},{Nout});',
        f'TURN({Sin},{Eout});',
        f'TURN({Sin},{Nout});',
        f'TURN({Sin},{Wout});',
         '',
         '//link-to',
        f'LINK-TO({Ein},{i});',
        f'LINK-TO({Nin},{i});',
        f'LINK-TO({Win},{i});',
        f'LINK-TO({Sin},{i});',
         '',
         '//link-from',
        f'LINK-FROM({i},{Eout});',
        f'LINK-FROM({i},{Nout});',
        f'LINK-FROM({i},{Wout});',
        f'LINK-FROM({i},{Sout});',
         '',
         '//timing constraints',
        f'PHASE-MIN({i},@WEST-EAST-LEFT) = {min_left};',
        f'PHASE-MIN({i},@WEST-EAST-THROUGH) = {min_through};',
        f'PHASE-MIN({i},@NORTH-SOUTH-LEFT) = {min_left};',
        f'PHASE-MIN({i},@NORTH-SOUTH-THROUGH) = {min_through};',
        f'PHASE-MAX({i},@WEST-EAST-LEFT) = {max};',
        f'PHASE-MAX({i},@WEST-EAST-THROUGH) = {max};',
        f'PHASE-MAX({i},@NORTH-SOUTH-LEFT) = {max};',
        f'PHASE-MAX({i},@NORTH-SOUTH-THROUGH) = {max};',
         '',
        f'PHASE-MIN({i},@ALL-RED) = {red};'
        f'PHASE-MIN({i},@ALL-RED2) = {red};'
        f'PHASE-MIN({i},@ALL-RED3) = {red};'
        f'PHASE-MIN({i},@ALL-RED4) = {red};'
        f'PHASE-MAX({i},@ALL-RED) = {red};'
        f'PHASE-MAX({i},@ALL-RED2) = {red};'
        f'PHASE-MAX({i},@ALL-RED3) = {red};'
        f'PHASE-MAX({i},@ALL-RED4) = {red};'
         '//green turns',
        f'GREEN({Ein},{Sout},@WEST-EAST-LEFT);',
        f'GREEN({Win},{Nout},@WEST-EAST-LEFT);',
        f'GREEN({Ein},{Wout},@WEST-EAST-THROUGH);',
        f'GREEN({Win},{Eout},@WEST-EAST-THROUGH);',
        f'GREEN({Nin},{Eout},@NORTH-SOUTH-LEFT);',
        f'GREEN({Sin},{Wout},@NORTH-SOUTH-LEFT);',
        f'GREEN({Nin},{Sout},@NORTH-SOUTH-THROUGH);',
        f'GREEN({Sin},{Nout},@NORTH-SOUTH-THROUGH);'
    ))

    # Add right turns on green
    right_turn_pairs = ((Ein,Nout),
                        (Nin,Wout),
                        (Win,Sout),
                        (Sin,Eout))
    phases = ('@WEST-EAST-LEFT',
              '@WEST-EAST-THROUGH',
              '@NORTH-SOUTH-LEFT',
              '@NORTH-SOUTH-THROUGH')

    nonfluents_str += newline_indent_str
    nonfluents_str += newline_indent_str.join( (f'GREEN({i},{j},{p});'
                                                for (i,j) in right_turn_pairs for p in phases) )

    # Optionally, add right turns on red
    if right_on_red:
        red_phases = ('@ALL-RED',
                      '@ALL-RED2',
                      '@ALL-RED3',
                      '@ALL-RED4')
        nonfluents_str += newline_indent_str
        nonfluents_str += newline_indent_str.join( (f'GREEN({i},{j},{p});'
                                                    for (i,j) in right_turn_pairs for p in red_phases) )
    return nonfluents_str


def generate_nema_4leg_intersection(i, Ein, Nout, Nin, Wout, Win, Sout, Sin, Eout,
                                    min_left, min_through, max, red):
    """Generates the non-fluents for a NEMA four-leg intersection,
    with left, through, and right movements"""

    nonfluents_str = newline_indent_str*2 + f'//intersection {i}'
    nonfluents_str += newline_indent_str + '//turns' + newline_indent_str
    nonfluents_str += newline_indent_str.join((
        f'TURN({Ein},{Nout});',
        f'TURN({Ein},{Wout});',
        f'TURN({Ein},{Sout});',
        f'TURN({Nin},{Wout});',
        f'TURN({Nin},{Sout});',
        f'TURN({Nin},{Eout});',
        f'TURN({Win},{Sout});',
        f'TURN({Win},{Eout});',
        f'TURN({Win},{Nout});',
        f'TURN({Sin},{Eout});',
        f'TURN({Sin},{Nout});',
        f'TURN({Sin},{Wout});',
         '',
         '//link-to',
        f'LINK-TO({Ein},{i});',
        f'LINK-TO({Nin},{i});',
        f'LINK-TO({Win},{i});',
        f'LINK-TO({Sin},{i});',
         '',
         '//link-from',
        f'LINK-FROM({i},{Eout});',
        f'LINK-FROM({i},{Nout});',
        f'LINK-FROM({i},{Wout});',
        f'LINK-FROM({i},{Sout});',
         '',
         '//timing constraints',
        f'PHASE-MIN({i},p0) = {min_left};',
        f'PHASE-MIN({i},p1) = {min_left};',
        f'PHASE-MIN({i},p2) = {min_left};',
        f'PHASE-MIN({i},p3) = {min_through};',
        f'PHASE-MIN({i},p4) = {min_left};',
        f'PHASE-MIN({i},p5) = {min_left};',
        f'PHASE-MIN({i},p6) = {min_left};',
        f'PHASE-MIN({i},p7) = {min_through};',
         '',
        f'PHASE-MAX({i},p0) = {max};',
        f'PHASE-MAX({i},p1) = {max};',
        f'PHASE-MAX({i},p2) = {max};',
        f'PHASE-MAX({i},p3) = {max};',
        f'PHASE-MAX({i},p4) = {max};',
        f'PHASE-MAX({i},p5) = {max};',
        f'PHASE-MAX({i},p6) = {max};',
        f'PHASE-MAX({i},p7) = {max};',
         '',
        f'PHASE-ALL-RED-DUR({i},p0) = {red};',
        f'PHASE-ALL-RED-DUR({i},p1) = {red};',
        f'PHASE-ALL-RED-DUR({i},p2) = {red};',
        f'PHASE-ALL-RED-DUR({i},p3) = {red};',
        f'PHASE-ALL-RED-DUR({i},p4) = {red};',
        f'PHASE-ALL-RED-DUR({i},p5) = {red};',
        f'PHASE-ALL-RED-DUR({i},p6) = {red};',
        f'PHASE-ALL-RED-DUR({i},p7) = {red};',
         '',
         '//green turns',
        f'GREEN({Ein},{Sout},p0); GREEN({Win},{Nout},p0);',
        f'GREEN({Ein},{Sout},p1); GREEN({Ein},{Wout},p1);',
        f'GREEN({Win},{Nout},p2); GREEN({Win},{Eout},p2);',
        f'GREEN({Ein},{Wout},p3); GREEN({Win},{Eout},p3);',
        f'GREEN({Nin},{Eout},p4); GREEN({Sin},{Wout},p4);',
        f'GREEN({Sin},{Wout},p5); GREEN({Sin},{Nout},p5);',
        f'GREEN({Nin},{Eout},p6); GREEN({Nin},{Sout},p6);',
        f'GREEN({Nin},{Sout},p7); GREEN({Sin},{Nout},p7);'
    ))

    # Add right turns on green
    right_turn_pairs = ((Ein,Nout),
                        (Nin,Wout),
                        (Win,Sout),
                        (Sin,Eout))
    phases = ('p0', 'p1', 'p2', 'p3', 'p4', 'p5', 'p6', 'p7')

    nonfluents_str += newline_indent_str
    nonfluents_str += newline_indent_str.join( (f'GREEN({i},{j},{p});'
                                                for (i,j) in right_turn_pairs for p in phases) )

    return nonfluents_str



def generate_grid(nrows,
                  ncols,
                  phasing_type,
                  ew_link_len=(200,50), #(a,b) parsed as Uniform(a-b,a+b)
                  ns_link_len=(200,50),
                  apply_off_grid_perturbation=False,
                  feeder_link_elongation_factor=1.5,
                  Vl=13.8,
                  inflow_rate_per_lane=(0.08,0.02), # parsed as Uniform(a-b,a+b)
                  satflow_per_lane=0.53,
                  num_lanes=4,
                  high_left_prob=0,
                  min_green_left=3,
                  min_green_through=7,
                  max_green=80,
                  all_red=4,
                  right_on_red=True,
                  instance_name=None,
                  horizon=200,
                  discount=1.0):
    """Generates a grid network.

    The inflow rates are sampled from a uniform random distribution,
    and so are the link lengths. The feeder links can be elongated
    to fit more vehicles and provide more information to the boundary
    lights.

    Typically, through movements are assumed to get (2/4) of the lanes,
    and left and right turns 1/4 each. The saturation flow rate for a
    movement is obtained by multiplying the sat. flow rate per lane by
    the assumed number of lanes.

    There is a fixed probability for a left-turn to have higher demand than the
    through movement (defaults to 0). In this case, the left turns are assumed
    to have (2/4) of the lanes and the through movements (1/4) of the lanes.
    """
    assert phasing_type in {'simple', 'nema'}, \
        '[netgen.generate_grid] Only "simple" and "nema" phasing types currently supported'

    # Sample link lengths uniformly
    # Optionally, make the feeder links longer to fit more vehicles
    feeder_elongation_ew, feeder_elongation_ns = np.ones(ncols+1), np.ones(nrows+1)
    feeder_elongation_ew[0] = feeder_elongation_ew[-1] = \
        feeder_elongation_ns[0] = feeder_elongation_ns[-1] = feeder_link_elongation_factor

    ew_lens = np.random.uniform(ew_link_len[0]-ew_link_len[1], ew_link_len[0]+ew_link_len[1], ncols+1)
    ns_lens = np.random.uniform(ns_link_len[0]-ns_link_len[1], ns_link_len[0]+ns_link_len[1], nrows+1)

    ew_lens *= feeder_elongation_ew
    ns_lens *= feeder_elongation_ns
    max_len = np.max(np.concatenate((ew_lens, ns_lens)))

    # Derive the X and Y coordinates of the intersections, sinks and sources
    Xs, Ys = np.zeros(ncols+2), np.zeros(nrows+2)
    Xs[1:] = np.cumsum(ew_lens)
    Ys[1:] = np.cumsum(ns_lens)
    Ys = np.flip(Ys) # Want Ys to decrease with increasing i to be consistent with
                     # cartesian coords
    Xs = np.tile(Xs, (nrows+2,1))
    Ys = np.tile(Ys[np.newaxis,:].transpose(), (1,ncols+2))
    coords = np.concatenate((Xs[:,:,np.newaxis],
                             Ys[:,:,np.newaxis]),
                             axis=2)
    if apply_off_grid_perturbation:
        coords = coords + np.sqrt(50) * np.random.normal(size=coords.shape)
    coords = np.round(coords)


    num_intersections = nrows*ncols
    num_bdry = 2*(nrows + ncols)
    N = num_intersections + num_bdry
    num_ts = int(np.ceil(max_len/Vl))+2


    intersection_names = tuple(f'i{i}' for i in range(num_intersections))
    t_names = tuple(f't{i}' for i in range(num_ts))


    inames = np.array(['EMPTY' for _ in range(N+4)]).reshape((nrows+2,ncols+2))

    for i in range(nrows+2):
        for j in range(ncols+2):

            if 0 < i < nrows+1 and 0 < j < ncols+1:
                # Traffic light
                inames[i,j] = f'i{(j-1) + (i-1)*ncols}'
            else:
                # Source/sink
                if 0 < j < ncols+1:
                    if i==0:
                        inames[i,j] = f's{j-1}'
                    elif i==(nrows+1):
                        inames[i,j] = f's{2*ncols + nrows - j}'
                if 0 < i < nrows+1:
                    if j==(ncols+1):
                        inames[i,j] = f's{ncols+(i-1)}'
                    elif j==0:
                        inames[i,j] = f's{num_bdry - i}'


    link_names = []
    link_lengths = []
    left_turns, through_turns, right_turns = [], [], []

    for i in range(1,nrows+1):
        for j in range(1,ncols+1):
            link_names.extend([
                f'l-{inames[i,j]}-{inames[i-1,j]}',
                f'l-{inames[i,j]}-{inames[i,j+1]}',
                f'l-{inames[i,j]}-{inames[i+1,j]}',
                f'l-{inames[i,j]}-{inames[i,j-1]}'])
            link_lengths.extend([
                dist(coords[i-1,j], coords[i,j]),
                dist(coords[i,j+1], coords[i,j]),
                dist(coords[i+1,j], coords[i,j]),
                dist(coords[i,j-1], coords[i,j])])

            left_turns.extend([
                f'l-{inames[i,j-1]}-{inames[i,j]},l-{inames[i,j]}-{inames[i-1,j]}',
                f'l-{inames[i-1,j]}-{inames[i,j]},l-{inames[i,j]}-{inames[i,j+1]}',
                f'l-{inames[i,j+1]}-{inames[i,j]},l-{inames[i,j]}-{inames[i+1,j]}',
                f'l-{inames[i+1,j]}-{inames[i,j]},l-{inames[i,j]}-{inames[i,j-1]}'])
            through_turns.extend([
                f'l-{inames[i,j-1]}-{inames[i,j]},l-{inames[i,j]}-{inames[i,j+1]}',
                f'l-{inames[i,j+1]}-{inames[i,j]},l-{inames[i,j]}-{inames[i,j-1]}',
                f'l-{inames[i-1,j]}-{inames[i,j]},l-{inames[i,j]}-{inames[i+1,j]}',
                f'l-{inames[i+1,j]}-{inames[i,j]},l-{inames[i,j]}-{inames[i-1,j]}'])
            right_turns.extend([
                f'l-{inames[i,j-1]}-{inames[i,j]},l-{inames[i,j]}-{inames[i+1,j]}',
                f'l-{inames[i+1,j]}-{inames[i,j]},l-{inames[i,j]}-{inames[i,j+1]}',
                f'l-{inames[i,j+1]}-{inames[i,j]},l-{inames[i,j]}-{inames[i-1,j]}',
                f'l-{inames[i-1,j]}-{inames[i,j]},l-{inames[i,j]}-{inames[i,j-1]}'])

    # Add missing source links
    source_link_names = []
    for j in range(1,ncols+1):
        new_links = [
            f'l-{inames[0,j]}-{inames[1,j]}',
            f'l-{inames[nrows+1,j]}-{inames[nrows,j]}']
        link_names.extend(new_links)
        source_link_names.extend(new_links)
        link_lengths.extend([
            dist(coords[1,j], coords[0,j]),
            dist(coords[nrows,j], coords[nrows+1,j]) ])

    for i in range(1,nrows+1):
        new_links = [
            f'l-{inames[i,0]}-{inames[i,1]}',
            f'l-{inames[i,ncols+1]}-{inames[i,ncols]}']
        link_names.extend(new_links)
        source_link_names.extend(new_links)
        link_lengths.extend([
            dist(coords[i,1], coords[i,0]),
            dist(coords[i,ncols], coords[i,ncols+1]) ])

    # Optionally, make some left turns have heavier demand than the
    # through movements
    high_left_turn = np.random.binomial(1, high_left_prob, size=len(left_turns))
    deltas = np.random.uniform(-0.1, 0.1, size=len(left_turns))

    satflow_rates, turn_probs = {}, {}
    for L, T, R, dp, high_left in zip(left_turns, through_turns, right_turns, deltas, high_left_turn):
        if high_left:
            high_turn, low_turn = L, T
        else:
            high_turn, low_turn = T, L

        turn_probs[high_turn] = 0.7 - dp
        turn_probs[low_turn] = 0.2 + dp
        turn_probs[R] = 1-turn_probs[high_turn]-turn_probs[low_turn]
        total_satflow = satflow_per_lane * num_lanes
        satflow_rates[high_turn] = 0.5 * total_satflow
        satflow_rates[low_turn] = 0.3 * total_satflow
        satflow_rates[R] = 0.2 * total_satflow

    inflow_lb = (inflow_rate_per_lane[0]-inflow_rate_per_lane[1]) * num_lanes
    inflow_ub = (inflow_rate_per_lane[0]+inflow_rate_per_lane[1]) * num_lanes
    arrival_rates = np.round( np.random.uniform(inflow_lb, inflow_ub, num_bdry),
                              2)

    if instance_name is None:
        instance_name = f'grid_{nrows}x{ncols}'


    instance_str = '\n'.join((
        f'non-fluents {instance_name} {{',
        f'    domain = BLX_model;',
        f'',
        f'    objects {{',
        f'        intersection : {{{", ".join(intersection_names)}}};',
        f'        link         : {{{", ".join(link_names)}}};'))

    if phasing_type == 'nema':
        instance_str += '\n'.join((
            '',
            f'        signal-phase : {{p0, p1, p2, p3, p4, p5, p6, p7}};',
            f'        action-token : {{a0, a1, a2, a3, a4}};'))

    instance_str += '\n'.join((
        '',
        f'        time         : {{{", ".join(t_names)}}};',
        f'    }};',
        f'',
        f'    non-fluents {{'))

    if phasing_type == 'nema':
        instance_str += newline_indent_str.join(('',) + tuple(
            f'INTERSECTION-INDEX({i}) = {ni};' for ni, i in enumerate(intersection_names)))
        instance_str += newline_indent_str.join((
            '',
            '',
            'ACTION-TOKEN-INDEX(a0) = 0;',
            'ACTION-TOKEN-INDEX(a1) = 1;',
            'ACTION-TOKEN-INDEX(a2) = 2;',
            'ACTION-TOKEN-INDEX(a3) = 3;',
            'ACTION-TOKEN-INDEX(a4) = 4;',
            ''
            'PHASE-INDEX(p0) = 0;',
            'PHASE-INDEX(p1) = 1;',
            'PHASE-INDEX(p2) = 2;',
            'PHASE-INDEX(p3) = 3;',
            'PHASE-INDEX(p4) = 4;',
            'PHASE-INDEX(p5) = 5;',
            'PHASE-INDEX(p6) = 6;',
            'PHASE-INDEX(p7) = 7;',
            ''))

    instance_str += newline_indent_str + '//sources'
    instance_str += newline_indent_str.join(('',)
        + tuple(f'SOURCE(l-{inames[i,0]}-{inames[i,1]});'
                + newline_indent_str
                + f'SOURCE(l-{inames[i,ncols+1]}-{inames[i,ncols]});'
                for i in range(1, nrows+1))
        + tuple(f'SOURCE(l-{inames[0,j]}-{inames[1,j]});'
                + newline_indent_str
                + f'SOURCE(l-{inames[nrows+1,j]}-{inames[nrows,j]});'
                for j in range(1, ncols+1)))

    instance_str += newline_indent_str + '//sinks'
    instance_str += newline_indent_str.join(('',)
        + tuple(f'SINK(l-{inames[i,1]}-{inames[i,0]});'
                + newline_indent_str
                + f'SINK(l-{inames[i,ncols]}-{inames[i,ncols+1]});'
                for i in range(1, nrows+1))
        + tuple(f'SINK(l-{inames[1,j]}-{inames[0,j]});'
                + newline_indent_str
                + f'SINK(l-{inames[nrows,j]}-{inames[nrows+1,j]});'
                for j in range(1, ncols+1)))


    if Vl != 13.8:
        instance_str += newline_indent_str + '//speeds'
        instance_str += newline_indent_str.join(('',) + tuple(f'SPEED({link}) = {Vl};' for link in link_names))

    if num_lanes != 4:
        instance_str += newline_indent_str + '//number of lanes'
        instance_str += newline_indent_str.join(('',) + tuple(f'Nl({link}) = {num_lanes};' for link in link_names))

    instance_str += newline_indent_str + '//satflow rates'
    instance_str += newline_indent_str.join(('',) + tuple(f'MU({k}) = {v};' for k,v in satflow_rates.items()))

    instance_str += newline_indent_str + '//turn probabilities'
    instance_str += newline_indent_str.join(('',) + tuple(f'BETA({k}) = {v};' for k,v in turn_probs.items()))

    instance_str += newline_indent_str + '//link lengths'
    instance_str += newline_indent_str.join(('',) + tuple(f'Dl({k}) = {v};' for k,v in zip(link_names, link_lengths)))

    instance_str += newline_indent_str + '//source arrival rates'
    instance_str += newline_indent_str.join(('',) + tuple(f'SOURCE-ARRIVAL-RATE({k}) = {v};' for k,v in zip(source_link_names, arrival_rates)))

    for i in range(1, nrows+1):
        for j in range(1, ncols+1):
            if phasing_type == 'simple':
                instance_str += generate_simple_4leg_intersection(
                                    inames[i,j],
                                    f'l-{inames[i,j+1]}-{inames[i,j]}', #Ein
                                    f'l-{inames[i,j]}-{inames[i-1,j]}', #Nout
                                    f'l-{inames[i-1,j]}-{inames[i,j]}', #Nin
                                    f'l-{inames[i,j]}-{inames[i,j-1]}', #Wout
                                    f'l-{inames[i,j-1]}-{inames[i,j]}', #Win
                                    f'l-{inames[i,j]}-{inames[i+1,j]}', #Sout
                                    f'l-{inames[i+1,j]}-{inames[i,j]}', #Sin
                                    f'l-{inames[i,j]}-{inames[i,j+1]}', #Eout
                                    min_left=min_green_left,
                                    min_through=min_green_through,
                                    max=max_green,
                                    red=all_red,
                                    right_on_red=right_on_red)
            elif phasing_type == 'nema':
                instance_str += generate_nema_4leg_intersection(
                                    inames[i,j],
                                    f'l-{inames[i,j+1]}-{inames[i,j]}', #Ein
                                    f'l-{inames[i,j]}-{inames[i-1,j]}', #Nout
                                    f'l-{inames[i-1,j]}-{inames[i,j]}', #Nin
                                    f'l-{inames[i,j]}-{inames[i,j-1]}', #Wout
                                    f'l-{inames[i,j-1]}-{inames[i,j]}', #Win
                                    f'l-{inames[i,j]}-{inames[i+1,j]}', #Sout
                                    f'l-{inames[i+1,j]}-{inames[i,j]}', #Sin
                                    f'l-{inames[i,j]}-{inames[i,j+1]}', #Eout
                                    min_left=min_green_left,
                                    min_through=min_green_through,
                                    max=max_green,
                                    red=all_red)

    if phasing_type == 'nema':
        instance_str += '\n        '.join(('', '//phase transitions',
            'TRANSITION(p0, a0) = 0;',
            'TRANSITION(p0, a1) = 1;',
            'TRANSITION(p0, a2) = 2;',
            'TRANSITION(p0, a3) = 3;',
            '',
            'TRANSITION(p1, a0) = 1;',
            'TRANSITION(p1, a1) = 3;',
            '',
            'TRANSITION(p2, a0) = 2;',
            'TRANSITION(p2, a1) = 3;',
            '',
            'TRANSITION(p3, a0) = 3;',
            'TRANSITION(p3, a1) = 4;',
            'TRANSITION(p3, a2) = 5;',
            'TRANSITION(p3, a3) = 6;',
            'TRANSITION(p3, a4) = 7;',
            '',
            'TRANSITION(p4, a0) = 4;',
            'TRANSITION(p4, a1) = 5;',
            'TRANSITION(p4, a2) = 6;',
            'TRANSITION(p4, a3) = 7;',
            '',
            'TRANSITION(p5, a0) = 5;',
            'TRANSITION(p5, a1) = 7;',
            '',
            'TRANSITION(p6, a0) = 6;',
            'TRANSITION(p6, a1) = 7;',
            '',
            'TRANSITION(p7, a0) = 7;',
            'TRANSITION(p7, a1) = 0;',
            'TRANSITION(p7, a2) = 1;',
            'TRANSITION(p7, a3) = 2;',
            'TRANSITION(p7, a4) = 3;'))

    instance_str += '\n        '.join(('', '// time-delay properties',
       f'TIME-HEAD(t0);',
       f'TIME-TAIL(t{num_ts-1});') +
        tuple(f'TIME-VAL(t{i}) = {i};' for i in range(num_ts)) +
        tuple(f'NEXT(t{i},t{i+1});' for i in range(num_ts-1)))


    instance_str += newline_indent_str + '//cartesian coordinates (for visualization)'
    instance_str += newline_indent_str.join(('',) + tuple(
        f'X({inames[i,j]}) = {coords[i,j,0]}; Y({inames[i,j]}) = {coords[i,j,1]};'
        for i in range(1,nrows+1) for j in range(1,ncols+1) ))

    instance_str += newline_indent_str + newline_indent_str.join((
        f'SOURCE-X(l-{inames[i,0]}-{inames[i,1]}) = {coords[i,0,0]}; SOURCE-Y(l-{inames[i,0]}-{inames[i,1]}) = {coords[i,0,1]};'
        + newline_indent_str
        + f'SOURCE-X(l-{inames[i,ncols+1]}-{inames[i,ncols]}) = {coords[i,ncols+1,0]}; SOURCE-Y(l-{inames[i,ncols+1]}-{inames[i,ncols]}) = {coords[i,ncols+1,1]};'
        + newline_indent_str
        + f'SINK-X(l-{inames[i,1]}-{inames[i,0]}) = {coords[i,0,0]}; SINK-Y(l-{inames[i,1]}-{inames[i,0]}) = {coords[i,0,1]};'
        + newline_indent_str
        + f'SINK-X(l-{inames[i,ncols]}-{inames[i,ncols+1]}) = {coords[i,ncols+1,0]}; SINK-Y(l-{inames[i,ncols]}-{inames[i,ncols+1]}) = {coords[i,ncols+1,1]};'
        for i in range(1,nrows+1) ))

    instance_str += newline_indent_str + newline_indent_str.join((
        f'SOURCE-X(l-{inames[0,j]}-{inames[1,j]}) = {coords[0,j,0]}; SOURCE-Y(l-{inames[0,j]}-{inames[1,j]}) = {coords[0,j,1]};'
        + newline_indent_str
        + f'SOURCE-X(l-{inames[nrows+1,j]}-{inames[nrows,j]}) = {coords[nrows+1,j,0]}; SOURCE-Y(l-{inames[nrows+1,j]}-{inames[nrows,j]}) = {coords[nrows+1,j,1]};'
        + newline_indent_str
        + f'SINK-X(l-{inames[1,j]}-{inames[0,j]}) = {coords[0,j,0]}; SINK-Y(l-{inames[1,j]}-{inames[0,j]}) = {coords[0,j,1]};'
        + newline_indent_str
        + f'SINK-X(l-{inames[nrows,j]}-{inames[nrows+1,j]}) = {coords[nrows+1,j,0]}; SINK-Y(l-{inames[nrows,j]}-{inames[nrows+1,j]}) = {coords[nrows+1,j,1]};'
        for j in range(1,ncols+1) ))


    instance_str += '\n'
    instance_str += '\n'.join((
        f'    }};',
        f'}}',
        f'',
        f'instance {instance_name} {{',
        f'    domain = BLX_model;',
        f'    non-fluents = {instance_name};',
        f'    max-nondef-actions = {num_intersections};',
        f'    horizon = {horizon};',
        f'    discount = {discount};',
        f'}}' ))

    instance_str += '\n'
    instance_str += '// Source link ids\n//'
    source_link_indices = [f'{id}={link_names.index(id)}' for id in source_link_names]
    instance_str += '\n//'.join(source_link_indices)

    return instance_str





if __name__ == '__main__':
    import argparse
    import os

    parser = argparse.ArgumentParser(description='Tool for automatically generating grid instances for the RDDL traffic domain')
    parser.add_argument('target_path', type=str, help='Path the generated rddl code will be saved to')
    parser.add_argument('-r', '--rows', type=int, help='Number of rows in the network', required=True)
    parser.add_argument('-c', '--cols', type=int, help='Number of columns in the network', required=True)
    parser.add_argument('-p', '--phasing', type=str, help='The phasing type. Either "simple" or "nema"', required=True)
    parser.add_argument('-f', '--force-overwrite', action='store_true', help='By default the generator will not overwrite existing files. With this argument, it will')
    parser.add_argument('-L', '--high-left-prob', default=0, help='Probability of having heavier demand on through than left from an approach')
    parser.add_argument('-T', '--horizon', type=int, default=200)
    parser.add_argument('--off-grid', action='store_true', help='Apply small perturbations to move the intersections off a perfect grid, making the network less symmetric')
    parser.add_argument('-n', '--instance-name', help='Name of instance')
    args = parser.parse_args()


    args.high_left_prob = float(args.high_left_prob)
    assert(0 <= args.high_left_prob <= 1)

    if os.path.isfile(args.target_path) and not args.force_overwrite:
        raise RuntimeError('[netgen.py] File with the requested path already exists. Pass a diffent path or add the -f argument to force overwrite')

    with open(args.target_path, 'w') as file:
        network = generate_grid(
            args.rows, args.cols,
            phasing_type=args.phasing,
            apply_off_grid_perturbation=args.off_grid,
            high_left_prob=args.high_left_prob,
            horizon=args.horizon,
            instance_name=args.instance_name)

        file.write(network)
    print(f'[netgen.py] Successfully generated the network instance RDDL file to {args.target_path}')



info = {
    'name': 'PowerGen_Continuous',
    'description': 'A simple continuous version of the power generation problem, loosely modeled on the problem of unit commitment.',
    'context': '',
    'tags': '',
    'viz': 'PowerGenViz.PowerGenVisualizer'
}


info = {
    'name': 'Knapsack',
    'description': 'Items of random weight and value are drawn, the goal is to place them into knapsacks of limited total weight while maximizing total value of all items.',
    'context': 'or',
    'tags': '',
    'viz': 'None'
}


info = {
    'name': 'HVAC',
    'description': 'Multi-zone and multi-heater HVAC control problem',
    'context': 'ippc2023',
    'tags': '',
    'viz': 'HVACViz.HVACVisualizer',
    'generator': 'HVACInstanceGenerator.HVACInstanceGenerator'
}


info = {
    'name': 'BinPacking',
    'description': 'Items of random weight are drawn, the goal is to place them into bins while minimizing the number of bins used and the total weight of each bin is within limits.',
    'context': 'or',
    'tags': '',
    'viz': 'None'
}


info = {
    'name': 'TriangleTireworld_POMDP',
    'description': 'In short, this problem was intended to be difficult for determinization/replanning approaches since the highest probability path to the goal is longer than other lower probability (but still possible) paths to the goal. This is the pomdp version',
    'context': 'ippc2014',
    'tags': '',
    'viz': 'None'
}



import numpy as np
from PIL import Image
import pygame
from pygame import gfxdraw

from pyRDDLGym.core.compiler.model import RDDLPlanningModel
from pyRDDLGym.core.visualizer.viz import BaseViz


# code comes from openai gym
class PendulumVisualizer(BaseViz):

    def __init__(self, model: RDDLPlanningModel, screen_dim=500, wait_time=100) -> None:
        self._model = model
        self._wait_time = wait_time
        self.screen_dim = screen_dim
    
    def init_canvas(self, figure_size):
        screen = pygame.Surface(figure_size)
        surf = pygame.Surface(figure_size)
        return screen, surf
        
    def convert2img(self, screen):
        data = np.transpose(np.array(pygame.surfarray.pixels3d(screen)), 
                            axes=(1, 0, 2))
        img = Image.fromarray(data)
        return img
    
    def render(self, state):
        screen, surf = self.init_canvas((self.screen_dim, self.screen_dim))
        surf.fill((255, 255, 255))
        
        bound = 2.2
        scale = self.screen_dim / (bound * 2)
        offset = self.screen_dim // 2
        
        # draw the arm
        rod_length = 1 * scale
        rod_width = 0.2 * scale
        l, r, t, b = 0, rod_length, rod_width / 2, -rod_width / 2
        coords = [(l, b), (l, t), (r, t), (r, b)]
        transformed_coords = []
        for c in coords:
            c = pygame.math.Vector2(c).rotate_rad(state['theta'] + np.pi / 2)
            c = (c[0] + offset, c[1] + offset)
            transformed_coords.append(c)
        gfxdraw.aapolygon(surf, transformed_coords, (204, 77, 77))
        gfxdraw.filled_polygon(surf, transformed_coords, (204, 77, 77))
        
        # draw the tip
        gfxdraw.aacircle(surf, offset, offset, int(rod_width / 2), (204, 77, 77))
        gfxdraw.filled_circle(
            surf, offset, offset, int(rod_width / 2), (204, 77, 77)
        )
        rod_end = (rod_length, 0)
        rod_end = pygame.math.Vector2(rod_end).rotate_rad(state['theta'] + np.pi / 2)
        rod_end = (int(rod_end[0] + offset), int(rod_end[1] + offset))
        gfxdraw.aacircle(
            surf, rod_end[0], rod_end[1], int(rod_width / 2), (204, 77, 77)
        )
        gfxdraw.filled_circle(
            surf, rod_end[0], rod_end[1], int(rod_width / 2), (204, 77, 77)
        )
        
        # draw the main axle
        gfxdraw.aacircle(surf, offset, offset, int(0.05 * scale), (0, 0, 0))
        gfxdraw.filled_circle(surf, offset, offset, int(0.05 * scale), (0, 0, 0))

        surf = pygame.transform.flip(surf, False, True)
        screen.blit(surf, (0, 0))
        pygame.time.wait(self._wait_time)
        img = self.convert2img(screen)
        del screen, surf        
        
        return img
    



info = {
    'name': 'Blackjack',
    'description': 'A card is randomly drawn from a deck and added to the player hand, the goal is to stop with the highest value of the hand without going over 21.',
    'context': 'arcade',
    'tags': '',
    'viz': 'None'
}


import os.path
import sys
import importlib.util
from typing import Any, Dict, List

from .error import (
    RDDLRepoInstanceNotExistError,
    RDDLRepoInstanceDuplicationError,
    RDDLRepoUnresolvedDependencyError
)

sys.path.append(os.path.join('..', 'rddlrepository'))

VIZ_BACKEND_PACKAGE_NAME = 'pyRDDLGym'
DOMAIN_NAME = 'domain.rddl'


class ProblemInfo:

    def __init__(self, problem_data: Dict[str, Any]) -> None:
        self.name = problem_data['name']
        self.desc = problem_data['description']
        self.loc = problem_data['location']
        self.instances = problem_data['instances']
        self.viz = problem_data['viz']

    def get_domain(self) -> str:
        path = os.path.join(self.loc, DOMAIN_NAME)
        return path

    def get_instance(self, num: str) -> str:
        if str(num) not in self.instances:
            instances = ' '.join(self.instances)
            raise RDDLRepoInstanceNotExistError(
                f'Domain <{self.name}> does not contain instance <{num}>, '
                f'must be one of: {instances}.')
        instance = f'instance{num}.rddl'
        path = os.path.join(self.loc, instance)
        return path

    def list_instances(self) -> List[str]:
        return self.instances

    def get_visualizer(self) -> str:
        if self.viz == 'None':
            return None

        spec = importlib.util.find_spec(VIZ_BACKEND_PACKAGE_NAME)
        if spec is None:
            raise RDDLRepoUnresolvedDependencyError(
                f'{VIZ_BACKEND_PACKAGE_NAME} is not installed: '
                f'it can be installed with \'pip install {VIZ_BACKEND_PACKAGE_NAME}\'.')

        path_to_viz = []
        p = os.path.split(self.loc)
        while p[1] != 'rddlrepository':
            path_to_viz.insert(0, p[1])
            p = os.path.split(p[0])
        path_to_viz.insert(0, p[1])
        path_to_viz = '.'.join(path_to_viz)
    
        viz = None
        viz_info = self.viz
        if viz_info:
            module, viz_class_name = viz_info.strip().split('.')
            viz_package_name = path_to_viz + '.' + module
            viz_package = __import__(viz_package_name, {}, {}, viz_class_name)
            viz = getattr(viz_package, viz_class_name)
        return viz
    
    def __str__(self) -> str:
        attr = self.__dict__
        values = [f'{name}: {value}' for name, value in self.__dict__.items()]
        return '\n'.join(values)
    
    def register_instance(self, num: str, rddl: str) -> None:
        path = os.path.join(self.loc, f'instance{num}.rddl')
        if os.path.exists(path):
            raise RDDLRepoInstanceDuplicationError(
                f'Instance <{num}> already exists in domain <{self.name}>.')
        
        with open(path, 'w') as instance_file:
            instance_file.write(rddl)               
        self.instances.append(num)
        
        print(f'Instance <{num}> was successfully registered in rddlrepository '
              f'for domain <{self.name}>.')
    



info = {
    'name': 'CooperativeRecon',
    'description': 'In this domain, the planner controls one or more planetary rovers that examine objects of interest in order to detect life and take a picture of it.',
    'context': 'ippc2018',
    'tags': '',
    'viz': 'None'
}



info = {
    'name': 'Reservoir_Continuous',
    'description': 'Continuous action version of management of the water level in interconnected reservoirs.',
    'context': '',
    'tags': '',
    'viz': 'ReservoirViz.ReservoirVisualizer'
}


class RDDLRepoEnvironmentNotExistError(ValueError):
    pass


class RDDLRepoInstanceNotExistError(ValueError):
    pass


class RDDLRepoDomainNotExistError(ValueError):
    pass


class RDDLRepoProblemDuplicationError(ValueError):
    pass


class RDDLRepoInstanceDuplicationError(ValueError):
    pass


class RDDLRepoContextDuplicationError(ValueError):
    pass


class RDDLRepoUnresolvedDependencyError(ImportError):
    pass


class RDDLRepoManifestEmptyError(ValueError):
    pass


class RDDLRepoContextNotExistError(ValueError):
    pass



import os
import random
from typing import Dict

from pyRDDLGym.Examples.InstanceGenerator import InstanceGenerator


class PowerGenInstanceGenerator(InstanceGenerator):
    
    def get_env_path(self) -> str:
        return os.path.join('PowerGen', 'Continuous')
    
    def get_domain_name(self) -> str:
        return 'power_gen'
    
    def sample_instance(self, params: Dict[str, object]) -> Dict[str, object]:
        num_gas = params['num_gas']
        num_nuclear = params['num_nuclear']
        num_solar = params['num_solar']
        num_plants = num_gas + num_nuclear + num_solar
        
        plants = [f'p{i + 1}' for i in range(num_plants)]        
        plant_types = ['gas'] * num_gas + ['nuclear'] * num_nuclear + ['solar'] * num_solar
        plant_params = [self._sample_plant(t) for t in plant_types]
        
        nonfluents = {}
        nonfluents[f'MIN-CONSUMPTION'] = 2.0 * params['demand_scale']
        nonfluents[f'DEMAND-EXP-COEF'] = 0.01 * params['demand_scale']
        
        able_demand = 0.0
        nonfluents[f'TEMP-VARIANCE'] = params['temp_variance']
        for plant, param in zip(plants, plant_params):
            able_demand += param['PROD-UNITS-MAX']
            for key, value in param.items():
                nonfluents[f'{key}({plant})'] = value
        
        def demand(t):
            return nonfluents[f'MIN-CONSUMPTION'] + nonfluents[f'DEMAND-EXP-COEF'] * (t - 11.7) ** 2
        
        max_demand = max(demand(-30.0), demand(+40.0))
        print(f'max demand = {max_demand}, capacity = {able_demand}')
        
        states = {}
        states[f'temperature'] = random.uniform(*params['temp_range'])
        
        return {
            'objects': {'plant': plants},
            'non-fluents': nonfluents,
            'init-states': states,
            'horizon': params['horizon'],
            'discount': params['discount'],
            'max-nondef-actions': 'pos-inf'
        }
    
    def _sample_plant(self, plant_type):
        if plant_type == 'gas':
            return self._sample_gas_fired_plant()
        elif plant_type == 'nuclear':
            return self._sample_nuclear_plant()
        elif plant_type == 'solar':
            return self._sample_solar_plant()
        else:
            raise Exception(f'Invalid plant type {plant_type}.')
        
    def _sample_gas_fired_plant(self):
        return {
            'PROD-UNITS-MIN': 1.0,
            'PROD-UNITS-MAX': 6.0,
            'TURN-ON-COST': 6.0,
            'PROD-CHANGE-PENALTY': 1.0,
            'COST-PER-UNIT': 4.0,
            'PROD-SHAPE': 1.0,
            'PROD-SCALE': 1e-7
        }
    
    def _sample_nuclear_plant(self):
        return {
            'PROD-UNITS-MIN': 2.0,
            'PROD-UNITS-MAX': 20.0,
            'TURN-ON-COST': 60.0,
            'PROD-CHANGE-PENALTY': 2.0,
            'COST-PER-UNIT': 1.0,
            'PROD-SHAPE': 1.0,
            'PROD-SCALE': 1e-3
        }
        
    def _sample_solar_plant(self):
        return {
            'PROD-UNITS-MIN': 1.0,
            'PROD-UNITS-MAX': 4.0,
            'TURN-ON-COST': 4.0,
            'PROD-CHANGE-PENALTY': 0.5,
            'COST-PER-UNIT': 3.0,
            'PROD-SHAPE': 1.0,
            'PROD-SCALE': 1.0
        }





info = {
    'name': 'MarsRover',
    'description': 'Multi Rover Navigation, where a group of agent needs to harvest mineral',
    'context': 'ippc2023',
    'tags': '',
    'viz': 'MarsRoverViz.MarsRoverVisualizer',
    'generator': 'MarsRoverInstanceGenerator.MarsRoverInstanceGenerator'
}


import copy
from datetime import datetime
import os
import importlib
import csv
from typing import Dict, List, Optional

from .error import (
    RDDLRepoDomainNotExistError,
    RDDLRepoProblemDuplicationError,
    RDDLRepoManifestEmptyError,
    RDDLRepoContextNotExistError,
    RDDLRepoContextDuplicationError
)
from .info import ProblemInfo

HEADER = ['name', 'description', 'location', 'instances', 'viz', 'context', 'tags']
manifest = 'manifest.csv'

PACKAGE_NAME = 'rddlrepository'
ARCHIVE_NAME = 'archive'
DOMAIN_NAME = 'domain.rddl'
INFO_NAME = '__init__.py'


class RDDLRepoManager:

    def __init__(self, rebuild: bool=False) -> None:
        self.archiver_dict = {}
        self.archive_by_context = {}
        
        self.manager_path = os.path.dirname(os.path.abspath(__file__))
        manifest_path = os.path.join(self.manager_path, manifest)
        if os.path.isfile(manifest_path) and not rebuild:
            try:
                self._load_repo()
            except:
                raise RDDLRepoManifestEmptyError(
                    'An error occurred while loading the repository manifest, '
                    'please re-run with rebuild=True.')
        else:
            self._build_repo()
    
    # ==========================================================================
    # GETTERS
    # ==========================================================================
    
    @staticmethod
    def _print_columns(values, cols=3, prefix=''):
        width = 1 + max(len(v) for v in values)
        result = prefix
        for (count, item) in enumerate(values, 1):
            result += item.ljust(width)
            if count % cols == 0 and count != len(values):
                result += '\n' + prefix         
        return result
        
    def list_problems(self) -> List[str]:
        if len(self.archiver_dict) == 0:
            raise RDDLRepoManifestEmptyError(
                'Repository manifest is empty: please re-run with rebuild=True.')        
        return list(self.archiver_dict.keys())

    def list_contexts(self) -> List[str]:
        if not self.archive_by_context:
            raise RDDLRepoManifestEmptyError(
                'Repository manifest is empty: please re-run with rebuild=True.')        
        return list(self.archive_by_context.keys())

    def list_problems_by_context(self, context: str) -> List[str]:
        info = self.archive_by_context.get(context, None)
        if info is None:
            valid_keys = list(self.archive_by_context.keys())
            raise RDDLRepoContextNotExistError(
                f'Context <{context}> does not exist in the repository, '
                f'must be one of:\n' + self._print_columns(valid_keys) + '\n')
        return copy.deepcopy(info)
    
    def get_problems_as_string(self) -> str:
        message = ''
        for context in self.list_contexts():
            message += context + ':'
            valid_keys = self.list_problems_by_context(context)
            message += '\n' +  self._print_columns(valid_keys, prefix='\t') + '\n'
        return message

    def get_problem(self, name: str) -> ProblemInfo:
        info = self.archiver_dict.get(name, None)
        if info is None:
            message = self.get_problems_as_string()
            raise RDDLRepoDomainNotExistError(
                f'Domain <{name}> does not exist in the repository, '
                f'must be one of:\n{message}')        
        return ProblemInfo(info)            
    
    # ==========================================================================
    # MANIFEST HANDLING
    # ==========================================================================
    
    def _build_repo(self) -> None:
        root_path = os.path.dirname(os.path.abspath(__file__))
        path_to_manifest = os.path.join(root_path, manifest)
        root_path = os.path.split(root_path)[0]
        archive_dir = os.path.join(root_path, ARCHIVE_NAME)
        start_char = len(archive_dir)

        # build the repo dictionary as first step to verify correctness and uniqueness
        for root, dirs, files in os.walk(archive_dir, topdown=False):
            dir = root[start_char:]
            if '__pycache__' in dirs:
                dirs.remove('__pycache__')
            if len(dirs) > 0:
                continue
            d = os.path.split(root_path)
            if d[1] == '__pycache__':
                continue
            
            if "__init__.py" in files:
                d = self._split_path_to_list(dir)
                module = PACKAGE_NAME + '.' + ARCHIVE_NAME + '.' + '.'.join(d)
                mymodule = importlib.import_module(module)
                context = mymodule.info['context']
                if context:
                    context = '_' + context
                name = mymodule.info['name'] + context
                
                if name in self.archiver_dict.keys():
                    raise RDDLRepoProblemDuplicationError(
                        f'Domain <{name}> already exists: problem names must be unique.')
                if DOMAIN_NAME not in files:
                    raise RDDLRepoDomainNotExistError(
                        f'Domain <{name}> does not have a {DOMAIN_NAME} file.')
                    
                instances = [fname[8:-5] for fname in files
                             if fname.startswith('instance') and fname.endswith('.rddl')]
                instances.sort(key=lambda x: int(x))
                context = mymodule.info['context']
                if context == '':
                    context = 'standalone'
                self.archiver_dict[name] = {
                    'name': name,
                    'description': mymodule.info['description'],
                    'location': root,
                    'instances': instances,
                    'viz': mymodule.info['viz'],
                    # 'context': mymodule.info['context'],
                    'context': context,
                    'tags': mymodule.info['tags']
                }

                # context = mymodule.info['context']
                # if context == '':
                #     context = 'standalone'
                self.archive_by_context.setdefault(context, []).append(name)

        # Generate manifest
        with open(path_to_manifest, 'w', newline='') as file:

            # write the csv header of the manifest
            writer = csv.writer(file, delimiter=',')
            writer.writerow(HEADER)

            # iterate through the dictionary
            for keys, values in self.archiver_dict.items():
                values_copy = copy.deepcopy(values)
                values_copy['instances'] = ','.join(values['instances'])
                values_copy['tags'] = ','.join(values['tags'])
                row = [values_copy[key] for key in HEADER]
                writer.writerow(row)

    def _load_repo(self) -> Dict:
        root_path = os.path.dirname(os.path.abspath(__file__))
        path_to_manifest = os.path.join(root_path, manifest)
        if not os.path.isfile(path_to_manifest):
            return {}

        self.archiver_dict = {}
        with open(path_to_manifest) as file:
            reader = csv.reader(file, delimiter=',')
            for i, row in enumerate(reader):
                if i > 0:
                    name, *entries = row
                    domain_info = dict(zip(HEADER[1:], entries))
                    domain_info['name'] = name
                    domain_info['instances'] = domain_info['instances'].split(',')                    
                    self.archiver_dict[name] = domain_info
                    if entries[4] == '':
                        context = 'standalone'
                    else:
                        context = entries[4]
                    self.archive_by_context.setdefault(context, []).append(name)
            return self.archiver_dict

    def _split_path_to_list(self, path):
        l = []
        a = os.path.split(path)
        while a[1] != '':
            l.insert(0, a[1])
            a = os.path.split(a[0])
        return l
    
    # ==========================================================================
    # REGISTRATION
    # ==========================================================================
    
    def register_context(self, context: str, refresh: bool=True) -> None:
        root_path = os.path.dirname(os.path.abspath(__file__))
        root_path = os.path.split(root_path)[0]
        context_dir = os.path.join(root_path, ARCHIVE_NAME, context)
        
        if context in self.archive_by_context or os.path.isdir(context_dir):
            raise RDDLRepoContextDuplicationError(
                f'Context <{context}> already exists.')
        
        os.mkdir(context_dir)
        open(os.path.join(context_dir, '__init__.py'), 'a').close()
        
        if refresh:
            self.archive_by_context[context] = []
        
        print(f'Context <{context}> was successfully registered in rddlrepository.')
    
    def register_domain(self, name: str, context: str, rddl: str,
                        desc: Optional[str]=None, viz: str='', refresh: bool=True) -> None:
        domains = self.list_problems_by_context(context)
        if name in domains:
            raise RDDLRepoProblemDuplicationError(
                f'Domain <{name}> already exists in context <{context}>.')
        
        root_path = os.path.dirname(os.path.abspath(__file__))
        root_path = os.path.split(root_path)[0]
        domain_dir = os.path.join(root_path, ARCHIVE_NAME, context, name)
        if os.path.isdir(domain_dir):
            raise RDDLRepoProblemDuplicationError(
                f'Domain <{name}> already exists in context <{context}>.')
        
        os.mkdir(domain_dir)
        
        if desc is None:
            desc = (f'User-defined domain with name {name} in context {context}, '
                    f'created on {datetime.today()}.')
        info = {'name': name, 'description': desc,
                'context': context, 'tags': '', 'viz': viz}
        
        with open(os.path.join(domain_dir, INFO_NAME), 'a') as info_file:
            info_file.write(f'info = {info}')
        with open(os.path.join(domain_dir, DOMAIN_NAME), 'a') as domain_file:
            domain_file.write(rddl)
        
        if refresh:
            self.archiver_dict = {}
            self.archive_by_context = {}
            self._build_repo()
            self._load_repo()
        
        print(f'Domain <{name}> was successfully registered in rddlrepository '
              f'with context <{context}>.')
        



import random
from typing import Dict

from pyRDDLGym.Examples.InstanceGenerator import InstanceGenerator


class MarsRoverInstanceGenerator(InstanceGenerator):
    
    def get_env_path(self) -> str:
        return 'MarsRover'
    
    def get_domain_name(self) -> str:
        return 'mars_rover_science_mission'
            
    def sample_instance(self, params: Dict[str, object]) -> Dict[str, object]:
        nm = params['num_minerals']
        nr = params['num_rovers']
        x1, x2 = params['location_bounds']
        y1, y2 = x1, x2
        a1, a2 = params['area_bounds']
        v1, v2 = params['value_bounds']
        
        minerals = [f'm{i + 1}' for i in range(nm)]
        rovers = [f'd{i + 1}' for i in range(nr)]
        
        nonfluents = {}
        for m in minerals:
            nonfluents[f'MINERAL-POS-X({m})'] = random.uniform(x1, x2)
            nonfluents[f'MINERAL-POS-Y({m})'] = random.uniform(y1, y2)
            nonfluents[f'MINERAL-AREA({m})'] = random.uniform(a1, a2)
            nonfluents[f'MINERAL-VALUE({m})'] = random.uniform(v1, v2)
        
        states = {}
        for r in rovers:
            states[f'pos-x({r})'] = random.uniform(x1, x2)
            states[f'pos-y({r})'] = random.uniform(y1, y2)
            states[f'vel-x({r})'] = 0.
            states[f'vel-y({r})'] = 0.
        
        return {
            'objects': {'mineral': minerals, 'rover': rovers},
            'non-fluents': nonfluents,
            'init-states': states,
            'horizon': params['horizon'],
            'discount': params['discount'],
            'max-nondef-actions': 'pos-inf'
        }



info = {
    'name': 'SkillTeaching_MDP',
    'description': 'In this domain, the agent is trying to teach a series of skills to a student through the use of hints and multiple choice questions.',
    'context': 'ippc2014',
    'tags': '',
    'viz': 'None'
}



info = {
    'name': 'Intruders_Continuous',
    'description': 'Continuous intruder detection problem on a unit square.',
    'context': '',
    'tags': '',
    'viz': 'IntrudersViz.IntrudersVisualizer'
}


import matplotlib.pyplot as plt
import numpy as np
from PIL import Image

from pyRDDLGym.core.compiler.model import RDDLPlanningModel
from pyRDDLGym.core.visualizer.viz import BaseViz


class WildfireVisualizer(BaseViz):

    def __init__(self, model: RDDLPlanningModel, dpi=50, fontsize=8) -> None:
        self._model = model
        self._nonfluents = model.ground_vars_with_values(model.non_fluents)
        self._objects = model.type_to_objects
        self._figure_size = None
        self._dpi = dpi
        self._fontsize = fontsize
        self._interval = 5

        self._nonfluent_layout = None
        self._state_layout = None
        self._fig, self._ax = None, None
        self._data = None
        self._img = None
    
    def build_nonfluents_layout(self): 
        targets = []        
        for k, v in self._nonfluents.items():
            var, objects = RDDLPlanningModel.parse_grounded(k)
            if var == 'TARGET':
                if v == True:
                    x, y = objects
                    targets.append((x, y))
        return {'targets':targets}

    def build_states_layout(self, state): 
        burning = []
        fuel = []
        for k, v in state.items():
            var, objects = RDDLPlanningModel.parse_grounded(k)
            if var == 'burning':
                if v == True:
                    x, y = objects
                    burning.append((x, y))
            if var == 'out-of-fuel':
                if v == True:
                    x, y = objects
                    fuel.append((x, y))        
        return {'burning':burning, 'out-of-fuel':fuel}

    def init_canvas_info(self):
        interval = self._interval
        x_list = self._objects['x_pos']
        y_list = self._objects['y_pos']
        grid_size = (int(len(x_list)), int(len(y_list)))

        grid_init_points = {}
        for x in x_list:
            for y in y_list:
                x_val = int(x[1:])
                y_val = int(y[1:])
                x_init = int((x_val - 1) * interval)
                y_init = int(grid_size[1] * interval - y_val * interval)
                grid_init_points[(x, y)] = (x_init, y_init)

        canvas_info = {
            'canvas_size': (grid_size[0] * interval, grid_size[1] * interval),
            'grid_init_points': grid_init_points
        }
        return canvas_info
        
    def init_canvas(self, figure_size, dpi):
        fig = plt.figure(figsize=figure_size, dpi=dpi)
        ax = plt.gca()
        plt.xlim([-0.5, figure_size[0] + 0.5])
        plt.ylim([-0.5, figure_size[1] + 0.5])
        plt.axis('scaled')
        return fig, ax

    def render_grid(self, pos, nonfluent_layout, state_layout):
        fig = self._fig
        ax = self._ax
        interval = self._interval
        init_x, init_y = self._canvas_info['grid_init_points'][pos]

        plt.text(init_x + 0.2,
                 init_y + 0.2,
                 str(pos),
                 color='black', fontsize=20, zorder=2)

        if pos in nonfluent_layout['targets']:
            if pos in state_layout['out-of-fuel']:
                color = "grey"
            else:
                color = "forestgreen"
            grid_rect = plt.Rectangle((init_x, init_y),
                                      interval,
                                      interval,
                                      fc=color, edgecolor='black',
                                      linewidth=2, zorder=0)
            plt.text(init_x + interval * 0.4,
                     init_y + interval * 0.4,
                     "Target",
                     color='black', fontsize=30, zorder=2)
        else:
            if pos in state_layout['out-of-fuel']:
                color = "lightgrey"
            else:
                color = "lawngreen"
            grid_rect = plt.Rectangle((init_x, init_y),
                                      interval,
                                      interval,
                                      fc=color, edgecolor='black',
                                      linewidth=2, zorder=0)
        
        ax.add_patch(grid_rect)
        
        if pos in state_layout['burning']:            
            piv_left = (init_x + 0.1, init_y)
            piv_right = (init_x + interval - 0.1, init_y)
            piv_top = (init_x + interval / 2 + 0.1, init_y + interval / 2)
            tri_fire = plt.Polygon([piv_left, piv_right, piv_top],
                                   closed=True, fc='tomato', zorder=1)
            ax.add_patch(tri_fire)

        return fig, ax

    def convert2img(self, fig, ax):        
        ax.set_position((0, 0, 1, 1))
        fig.canvas.draw()
        data = np.frombuffer(fig.canvas.buffer_rgba(), dtype=np.uint8)
        data = data.reshape(fig.canvas.get_width_height()[::-1] + (4,))
        data = data[:, :, :3]
        img = Image.fromarray(data)
        self._data = data
        self._img = img
        return img

    def fig2npa(self, fig):
        data = np.frombuffer(fig.canvas.buffer_rgba(), dtype=np.uint8)
        data = data.reshape(fig.canvas.get_width_height()[::-1] + (4,))
        data = data[:, :, :3]
        return data

    def render(self, state):
        self.states = state

        self._nonfluent_layout = self.build_nonfluents_layout()
        self._state_layout = self.build_states_layout(state)
        self._canvas_info = self.init_canvas_info()
        self._figure_size = self._canvas_info['canvas_size']
        self._fig, self._ax = self.init_canvas(self._figure_size, self._dpi)

        for pos in self._canvas_info['grid_init_points'].keys(): 
            self.render_grid(pos, self._nonfluent_layout, self._state_layout)

        img = self.convert2img(self._fig, self._ax)
            
        self._ax.cla()
        plt.cla()
        plt.close()

        return img




import numpy as np
from PIL import Image
import pygame
from pygame import gfxdraw

from pyRDDLGym.core.compiler.model import RDDLPlanningModel
from pyRDDLGym.core.visualizer.viz import BaseViz


class ReacherVisualizer(BaseViz):

    def __init__(self, model: RDDLPlanningModel, screen_dim=500, wait_time=100) -> None:
        self._model = model
        self._wait_time = wait_time
        self.screen_dim = screen_dim
        self._nonfluents = model.ground_vars_with_values(model.non_fluents)
    
    def init_canvas(self, figure_size):
        screen = pygame.Surface(figure_size)
        surf = pygame.Surface(figure_size)
        return screen, surf
        
    def convert2img(self, screen):
        data = np.transpose(np.array(pygame.surfarray.pixels3d(screen)), 
                            axes=(1, 0, 2))
        img = Image.fromarray(data)
        return img
    
    def render(self, state):
        segments = self._model.type_to_objects['segment']
        xpos = [state['tip-x___' + s] for s in segments]
        ypos = [state['tip-y___' + s] for s in segments]
        lengths = [self._nonfluents['LENGTH___' + s] for s in segments]
        bound = sum(lengths)
        
        screen, surf = self.init_canvas((self.screen_dim, self.screen_dim))
        surf.fill((255, 255, 255))
        scale = self.screen_dim / (bound * 2)
        offset = self.screen_dim // 2
        
        # draw the segments
        x, y = 0, 0
        for x2, y2 in zip(xpos, ypos):
            angle = np.arctan2(y2 - y, x2 - x)
            l = np.sqrt((x2 - x) ** 2 + (y2 - y) ** 2)
            rod_length = l * scale
            rod_width = 0.05 * scale
            l, r, t, b = 0, rod_length, rod_width / 2, -rod_width / 2
            coords = [(l, b), (l, t), (r, t), (r, b)]
            transformed_coords = []
            for c in coords:
                c = pygame.math.Vector2(c).rotate_rad(angle)
                c = (c[0] + offset + x * scale, c[1] + offset + y * scale)
                transformed_coords.append(c)
            gfxdraw.aapolygon(surf, transformed_coords, (204, 77, 77))
            gfxdraw.filled_polygon(surf, transformed_coords, (204, 77, 77))
            px = int(offset + x * scale)
            py = int(offset + y * scale)
            gfxdraw.aacircle(surf, px, py, int(rod_width / 2), (77, 77, 204))
            gfxdraw.filled_circle(surf, px, py, int(rod_width / 2), (77, 77, 204))
            x, y = x2, y2
        px = int(offset + x * scale)
        py = int(offset + y * scale)
        gfxdraw.aacircle(surf, px, py, int(rod_width / 2), (77, 77, 204))
        gfxdraw.filled_circle(surf, px, py, int(rod_width / 2), (77, 77, 204))
        
        # draw the target
        tx = self._model.non_fluents['TARGET-X']
        ty = self._model.non_fluents['TARGET-Y']
        px = int(offset + tx * scale)
        py = int(offset + ty * scale)

        gfxdraw.aacircle(surf, px, py, int(0.05 * scale / 2), (77, 204, 77))
        gfxdraw.filled_circle(surf, px, py, int(0.05 * scale / 2), (77, 204, 77))
        
        surf = pygame.transform.flip(surf, False, True)
        screen.blit(surf, (0, 0))
        pygame.time.wait(self._wait_time)
        img = self.convert2img(screen)
        del screen, surf        
        
        return img
    



info = {
    'name': 'UAV',
    'description': 'Continuous action space version of multi-UAV problem where a group of UAVs have to reach goal positions in the 3d Space',
    'context': 'ippc2023',
    'tags': '',
    'viz': 'UAVsViz.UAVsVisualizer',
    'generator': 'UAVInstanceGenerator.UAVInstanceGenerator'
}


info = {
    'name': 'Wildfire_POMDP',
    'description': 'A boolean version of the wildfire fighting domain. This is the pomdp version',
    'context': 'ippc2014',
    'tags': '',
    'viz': 'None'
}



info = {
    'name': 'TrafficBLX_ComplexPhases',
    'description': 'BLX/QTM traffic signal control model with a generic phasing scheme. The goal is to control traffic lights to minimize total travel time.',
    'context': '',
    'tags': '',
    'viz': 'TrafficViz.TrafficVisualizer'
}



info = {
    'name': 'CartPole_Discrete',
    'description': 'A simple continuous state discrete action MDP for the classical cart-pole system by Rich Sutton, with actions that describe the direction of the force applied to the cart.',
    'context': 'gym',
    'tags': '',
    'viz': 'CartPoleViz.CartPoleVisualizer'
}


import matplotlib.pyplot as plt
import numpy as np
from PIL import Image
from typing import Optional

from pyRDDLGym.core.compiler.model import RDDLPlanningModel
from pyRDDLGym.core.visualizer.viz import BaseViz


class HVACVisualizer(BaseViz):

    def __init__(self, model: RDDLPlanningModel,
                 grid_size: Optional[int] = [50, 50],
                 resolution: Optional[int] = [500, 500]) -> None:
        self._model = model
        self._states = model.ground_vars_with_values(model.state_fluents)
        self._nonfluents = model.ground_vars_with_values(model.non_fluents)
        self._objects = model.type_to_objects
        self._grid_size = grid_size
        self._resolution = resolution
        self._interval = 15

        self._object_layout = None
        self._canvas_info = None
        self._render_frame = None
        self._data = None
        self._fig = None
        self._ax = None

    def build_object_layout(self) -> dict:
        zone_vol = {o: None for o in self._objects['zone']}
        heater_vol = {o: None for o in self._objects['heater']}
        adj_zone = {o: [] for o in self._objects['zone']}
        adj_heater = {o: [] for o in self._objects['heater']}
        temp_zone = {o: None for o in self._objects['zone']}
        temp_heater = {o: None for o in self._objects['heater']}

        # add none-fluents
        for k, v in self._nonfluents.items():
            var, objects = RDDLPlanningModel.parse_grounded(k)
            if var == 'ZONE_VOL':
                zone_vol[objects[0]] = v
            elif var == 'HEATER-VOL':
                heater_vol[objects[0]] = v
            elif var == 'ADJ_ZONES':
                if v == True:
                    adj_zone[objects[0]].append(objects[1])
            elif var == 'ADJ_HEATER':
                if v == True:
                    adj_heater[objects[0]].append(objects[1])

        # add states
        for k, v in self._states.items():
            var, objects = RDDLPlanningModel.parse_grounded(k)
            if var == 'temp-zone':
                temp_zone[objects[0]] = v
            elif var == 'temp-heater':
                temp_heater[objects[0]] = v

        temp_zone_max = self._nonfluents['TEMP-ZONE-MAX']
        temp_zone_min = self._nonfluents['TEMP-ZONE-MIN']
        temp_out = self._nonfluents['TEMP-OUT']

        object_layout = {'temp-zone-max': temp_zone_max, 
                         'temp-zone-min': temp_zone_min,
                         'temp-out': temp_out,
                         'zone-vol': zone_vol,
                         'heater-vol': heater_vol,
                         'temp-zone': temp_zone,
                         'temp-heater': temp_heater,
                         'adj-zone': adj_zone,
                         'adj-heater': adj_heater}
        return object_layout

    def init_canvas_info(self):
        interval = self._interval

        zone_list = self._objects['zone']
        heater_list = self._objects['heater']

        room_grid_size = (max(len(zone_list), len(heater_list)),
                          max(len(zone_list), len(heater_list)) + 1)

        room_center = (room_grid_size[0] * self._interval / 2,
                       (room_grid_size[1] - 1) * self._interval / 2 + interval)

        start_points_zone = []
        for i in range(room_grid_size[0]):
            for j in range(room_grid_size[1]):
                start_points_zone.append((i * interval, j * interval))

        start_points_heater = []
        for i in range(room_grid_size[0]):
            start_points_heater.append((i * interval, 0))

        # rank start point by distance from center
        start_points_zone.sort(
            key=lambda x: abs(x[0] + interval / 2 - room_center[0]) + \
                          abs(x[1] + interval / 2 - room_center[1]))
        start_points_heater.sort(
            key=lambda x: abs(x[0] + interval / 2 - room_center[0]) + \
                          abs(x[1] + interval / 2 - room_center[1]))

        # rank by number of adj rooms
        zone_list.sort(reverse=True,
                       key=lambda x: len(self._object_layout['adj-zone'][x]))
        heater_list.sort(reverse=True,
                         key=lambda x: len(self._object_layout['adj-heater'][x]))

        zone_init_points = {zone_list[i]: start_points_zone[i]
                            for i in range(len(zone_list))}
        heater_init_points = {heater_list[i]: start_points_heater[i]
                              for i in range(len(heater_list))}

        canvas_info = {
            'canvas_size': (room_grid_size[0] * interval,
                            room_grid_size[1] * interval),
            'zone_init_points': zone_init_points,
            'heater_init_points': heater_init_points
        }
        return canvas_info

    def render_conn(self):
        ax = self._ax
        adj_zone = self._object_layout['adj-zone']
        adj_heater = self._object_layout['adj-heater']
        zone_init_points = self._canvas_info['zone_init_points']
        heater_init_points = self._canvas_info['heater_init_points']

        interval = self._interval * 2 / 3

        for k, v in adj_zone.items():
            from_point = (zone_init_points[k][0] + interval / 2,
                          zone_init_points[k][1] + interval / 2)
            for p in v:
                to_point = (zone_init_points[p][0] + interval / 2,
                            zone_init_points[p][1] + interval / 2)
                line = plt.Line2D((from_point[0], to_point[0]),
                                  (from_point[1], to_point[1]),
                                  ls='-', color='black', lw=2, zorder=1)
                ax.add_line(line)
        for k, v in adj_heater.items():
            from_point = (heater_init_points[k][0] + interval / 2,
                          heater_init_points[k][1] + interval / 2)
            for p in v:
                to_point = (zone_init_points[p][0] + interval / 2,
                            zone_init_points[p][1] + interval / 2)
                line = plt.Line2D((from_point[0], to_point[0]),
                                  (from_point[1], to_point[1]),
                                  ls='-', color='red', lw=2, zorder=1)
                ax.add_line(line)

    def render_zone(self, zone: str) -> tuple:
        ax = self._ax

        curr_z = zone
        vol_ratio = 1

        init_x, init_y = self._canvas_info['zone_init_points'][curr_z]
        interval = self._interval * 2 / 3 * vol_ratio

        temp = self._object_layout['temp-zone'][curr_z]
        upper_bound = self._object_layout['temp-zone-max']
        lower_bound = self._object_layout['temp-zone-min']

        upL = [init_x, upper_bound / 100 * interval + init_y]
        upR = [init_x + interval, upper_bound / 100 * interval + init_y]
        lowL = [init_x, lower_bound / 100 * interval + init_y]
        lowR = [init_x + interval, lower_bound / 100 * interval + init_y]

        background_rect = plt.Rectangle((init_x, init_y),
                                        interval,
                                        interval,
                                        fc='lightgrey', alpha=1, zorder=3)
        temp_rect = plt.Rectangle((init_x, init_y), interval, temp / 100 * interval, 
                                  fc='crimson', zorder=3)
        line_up = plt.Line2D((upL[0], upR[0]), (upL[1], upR[1]), 
                             ls='--', color='orange', lw=0.5, zorder=3)
        line_low = plt.Line2D((lowL[0], lowR[0]), (lowL[1], lowR[1]), 
                              ls='--', color='lime', lw=0.5, zorder=4)

        ax.add_patch(background_rect)
        ax.add_patch(temp_rect)
        ax.add_line(line_up)
        ax.add_line(line_low)

        plt.text(init_x + interval * 0.8, init_y + interval * 0.8, "%s" % curr_z, 
                 color='black', fontsize=5, zorder=5)

    def render_heater(self, heater: str) -> tuple:
        ax = self._ax

        curr_h = heater
        vol_ratio = 1

        init_x, init_y = self._canvas_info['heater_init_points'][curr_h]
        interval = self._interval * 2 / 3 * vol_ratio

        temp = self._object_layout['temp-heater'][curr_h]

        background_rect = plt.Rectangle((init_x, init_y), interval, interval,
                                        fc='lightcoral', alpha=1, zorder=3)
        temp_rect = plt.Rectangle((init_x, init_y), interval, temp / 100 * interval, 
                                  fc='crimson', zorder=3)

        ax.add_patch(background_rect)
        ax.add_patch(temp_rect)

        plt.text(init_x + interval * 0.8, init_y + interval * 0.8, "%s" % curr_h, 
                 color='black', fontsize=5, zorder=5)

    def fig2npa(self, fig):
        data = np.frombuffer(fig.canvas.buffer_rgba(), dtype=np.uint8)
        data = data.reshape(fig.canvas.get_width_height()[::-1] + (4,))
        data = data[:, :, :3]
        return data

    def render(self, state) -> np.ndarray:
        self._object_layout = self.build_object_layout()
        self._states = state
        self._canvas_info = self.init_canvas_info()

        canvas_size = self._canvas_info['canvas_size']
        self._fig = plt.figure(figsize=(canvas_size[0] / 10, canvas_size[1] / 10),
                               dpi=200)
        self._ax = plt.gca()

        plt.xlim([-canvas_size[0] / 5, canvas_size[0]])
        plt.ylim([-canvas_size[1] / 5, canvas_size[1]])
        plt.axis('scaled')
        plt.axis('off')

        for z in self._objects['zone']:
            self.render_zone(z)
        for h in self._objects['heater']:
            self.render_heater(h)

        self.render_conn()

        img = self.convert2img(self._fig, self._ax)

        self._ax.cla()
        plt.cla()
        plt.close()

        return img

    def convert2img(self, fig, ax):
        fig.canvas.draw()
        data = np.frombuffer(fig.canvas.buffer_rgba(), dtype=np.uint8)
        data = data.reshape(fig.canvas.get_width_height()[::-1] + (4,))
        data = data[:, :, :3]
        img = Image.fromarray(data)
        self._data = data
        self._img = img
        return img



import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import numpy as np
from PIL import Image
from typing import Optional

from pyRDDLGym.core.compiler.model import RDDLPlanningModel
from pyRDDLGym.core.visualizer.viz import BaseViz


class ReservoirVisualizer(BaseViz):

    def __init__(self, model: RDDLPlanningModel,
                 grid_size: Optional[int]=[50, 50],
                 resolution: Optional[int]=[500, 500]) -> None:
        self._model = model
        self._states = model.ground_vars_with_values(model.state_fluents)
        self._nonfluents = model.ground_vars_with_values(model.non_fluents)
        self._objects = model.type_to_objects
        self._grid_size = grid_size
        self._resolution = resolution
        self._interval = 15

        self._object_layout = None
        self._canvas_info = None
        self._render_frame = None
        self._data = None
        self._fig = None
        self._ax = None

    def build_object_layout(self) -> dict:
        max_res_cap = {o: None for o in self._objects['reservoir']}
        upper_bound = {o: None for o in self._objects['reservoir']}
        lower_bound = {o: None for o in self._objects['reservoir']}
        rain_shape = {o: None for o in self._objects['reservoir']}
        rain_scale = {o: None for o in self._objects['reservoir']}
        downstream = {o: [] for o in self._objects['reservoir']}
        sink_res = {o: None for o in self._objects['reservoir']}
        rlevel = {o: None for o in self._objects['reservoir']}

        # add none-fluents
        for k, v in self._nonfluents.items():
            var, objects = RDDLPlanningModel.parse_grounded(k)
            if var == 'TOP_RES':
                max_res_cap[objects[0]] = v
            elif var == 'MAX_LEVEL':
                upper_bound[objects[0]] = v
            elif var == 'MIN_LEVEL':
                lower_bound[objects[0]] = v
            elif var == 'RAIN_VAR':
                rain_shape[objects[0]] = v
            elif var == 'RES_CONNECT':
                if v == True:
                    downstream[objects[0]].append(objects[1])
            elif var == 'CONNECTED_TO_SEA':
                sink_res[objects[0]] = v

        # # add states
        for k, v in self._states.items():
            var, objects = RDDLPlanningModel.parse_grounded(k)
            if var == 'rlevel':
                rlevel[objects[0]] = v

        # adding defaults
        for o in self._objects['reservoir']:
            if max_res_cap[o] == None:
                max_res_cap[o] = self._nonfluents['TOP_RES']
            if upper_bound[o] == None:
                upper_bound[o] = self._nonfluents['MAX_LEVEL']
            if lower_bound[o] == None:
                lower_bound[o] = self._nonfluents['MIN_LEVEL']
            if rain_shape[o] == None:
                rain_shape[o] = self._nonfluents['RAIN_VAR']
            if rlevel[o] == None:
                rlevel[o] = self._states['rlevel']
            if sink_res[o] == None:
                sink_res[o] = False

        object_layout = {
            'rain_shape': rain_shape, 'max_res_cap': max_res_cap, 'rlevel': rlevel,
            'upper_bound': upper_bound, 'lower_bound': lower_bound, 'downstream': downstream,
            'sink_res': sink_res
        }
        return object_layout

    def init_canvas_info(self):
        interval = self._interval
        objects_set = set(self._objects['reservoir'])
        sink_res_set = set([k for k, v in self._object_layout['sink_res'].items() 
                            if v == True])

        all_child_set = []
        for i in self._object_layout['downstream'].values():
            all_child_set += i
        all_child_set = set(all_child_set) - sink_res_set

        all_root_set = objects_set - all_child_set - sink_res_set

        level_list = [list(all_root_set)]
        visited_nodes = objects_set.copy() - all_child_set

        top_set = all_root_set
        bot_set = set()

        while len(visited_nodes) < len(objects_set):
            for i in top_set:
                for j in self._object_layout['downstream'][i]:
                    bot_set.add(j)
                    visited_nodes.add(j)
            level_list.append(list(bot_set))
            top_set = bot_set
            bot_set = set()

        level_list.append(list(sink_res_set))

        row_num = len(level_list)
        col_num = max(len(i) for i in level_list)

        canvas_size = (col_num * interval, row_num * interval)
        init_points = {}
        for i in range(len(level_list)):
            for j in range(len(level_list[i])):
                init_x = 0 + interval * j
                init_y = canvas_size[1] - interval * (i + 1)
                init_points[level_list[i][j]] = (init_x, init_y)
                
        canvas_info = {'canvas_size': canvas_size, 'init_points': init_points}
        return canvas_info

    def render_conn(self):
        ax = self._ax
        downstream = self._object_layout['downstream']
        init_points = self._canvas_info['init_points']
        interval = self._interval * 2 / 3
        for k, v in downstream.items():
            top_point = (init_points[k][0] + interval / 2, init_points[k][1])
            for p in v:
                bot_point = (init_points[p][0] + interval / 2, init_points[p][1] + interval * 1.25)
                style = mpatches.ArrowStyle('Fancy', head_length=2, head_width=2, tail_width=0.01)
                arrow = mpatches.FancyArrowPatch(top_point, bot_point, arrowstyle=style, color='k')
                ax.add_patch(arrow)

    def render_res(self, res: str) -> tuple:
        fig = self._fig
        ax = self._ax
        interval = self._interval * 2 / 3
        curr_t = res
        init_x, init_y = self._canvas_info['init_points'][curr_t]

        rlevel = self._object_layout['rlevel'][curr_t]
        rain_shape = self._object_layout['rain_shape'][curr_t]

        max_res_cap = self._object_layout['max_res_cap'][curr_t]
        upper_bound = self._object_layout['upper_bound'][curr_t]
        lower_bound = self._object_layout['lower_bound'][curr_t]

        sink_res = self._object_layout['sink_res'][curr_t]

        maxL = [init_x, max_res_cap / 150 * interval + init_y]
        maxR = [init_x + interval, max_res_cap / 150 * interval + init_y]
        upL = [init_x, upper_bound / 150 * interval + init_y]
        upR = [init_x + interval, upper_bound / 150 * interval + init_y]
        lowL = [init_x, lower_bound / 150 * interval + init_y]
        lowR = [init_x + interval, lower_bound / 150 * interval + init_y]

        line_max = plt.Line2D((maxL[0], maxR[0]), (maxL[1], maxR[1]), 
                              ls='-', color='black', lw=0.25)
        line_up = plt.Line2D((upL[0], upR[0]), (upL[1], upR[1]), 
                             ls='--', color='orange', lw=0.25)
        line_low = plt.Line2D((lowL[0], lowR[0]), (lowL[1], lowR[1]), 
                              ls='--', color='orange', lw=0.25)
        lineL = plt.Line2D((init_x, init_x), (init_y, init_y + interval), 
                           color='black', lw=0.5)
        lineR = plt.Line2D((init_x + interval, init_x + interval), (init_y, init_y + interval), 
                           color='black', lw=0.5)
        lineB = plt.Line2D((init_x, init_x + interval), (init_y, init_y), 
                           color='black', lw=0.5)

        water_rect = plt.Rectangle((init_x, init_y), interval, rlevel / 120 * interval, 
                                   fc='royalblue')
        res_rect = plt.Rectangle((init_x, init_y), interval, interval, 
                                 fc='lightgrey', alpha=0.5)
        shape_rect = plt.Rectangle((init_x, init_y + interval), interval, interval / 4, 
                                   fc='darkcyan', alpha=np.clip(rain_shape / 50, 0, 1))

        ax.add_line(line_max)
        ax.add_line(line_up)
        ax.add_line(line_low)
        ax.add_line(lineL)
        ax.add_line(lineR)
        ax.add_line(lineB)

        lineU = plt.Line2D((init_x + interval * 1.25, init_x + interval * 1.25), 
                           (init_y, init_y + interval),
                           color='black', lw=1)
        lineD = plt.Line2D((init_x + interval, init_x + interval * 1.25), (init_y, init_y), 
                           color='black', lw=1)

        if sink_res:
            land_shape = plt.Rectangle((init_x + interval, init_y), interval / 4, interval, 
                                       fc='royalblue')
        else:
            land_shape = plt.Rectangle((init_x + interval, init_y), interval / 4, interval, 
                                       fc='darkgoldenrod')

        plt.text(init_x + interval * 1.1, init_y + interval * 1.1, "%s" % curr_t, 
                 color='black', fontsize=5)

        ax.add_patch(water_rect)
        ax.add_patch(res_rect)
        # ax.add_patch(scale_rect)
        ax.add_patch(shape_rect)

        ax.add_line(line_max)
        ax.add_line(line_up)
        ax.add_line(line_low)
        ax.add_line(lineL)
        ax.add_line(lineR)
        ax.add_line(lineB)
        ax.add_line(lineU)
        ax.add_line(lineD)

        ax.add_patch(land_shape)

        return fig, ax

    def fig2npa(self, fig):
        data = np.frombuffer(fig.canvas.buffer_rgba(), dtype=np.uint8)
        data = data.reshape(fig.canvas.get_width_height()[::-1] + (4,))
        data = data[:, :, :3]
        return data

    def render(self, state) -> np.ndarray:
        self._states = state
        self._object_layout = self.build_object_layout()
        self._canvas_info = self.init_canvas_info()
        canvas_size = self._canvas_info['canvas_size']
        self._fig = plt.figure(figsize=(canvas_size[0] / 10, canvas_size[1] / 10), dpi=200)
        self._ax = plt.gca()

        plt.xlim([0, canvas_size[0]])
        plt.ylim([0, canvas_size[1]])
        plt.axis('scaled')
        plt.axis('off')

        for res in self._objects['reservoir']:
            curr_t = res
            fig, ax = self.render_res(curr_t)
        self.render_conn()
        img = self.convert2img(self._fig, self._ax)

        self._ax.cla()
        plt.cla()
        plt.close()

        return img

    def convert2img(self, fig, ax):
        fig.canvas.draw()
        data = np.frombuffer(fig.canvas.buffer_rgba(), dtype=np.uint8)
        data = data.reshape(fig.canvas.get_width_height()[::-1] + (4,))
        data = data[:, :, :3]
        img = Image.fromarray(data)
        self._data = data
        self._img = img
        return img




import os
import random
from typing import Dict

from pyRDDLGym.Examples.InstanceGenerator import InstanceGenerator


class ReservoirInstanceGenerator(InstanceGenerator):
    
    def get_env_path(self) -> str:
        return os.path.join('Reservoir', 'Continuous')
    
    def get_domain_name(self) -> str:
        return 'reservoir_control_cont'
    
    def _generate_random_dag(self, n, maxedges):
        nodes = list(range(n))
        random.shuffle(nodes)
        graph = {node: set() for node in nodes}
        for i in range(n - 2, -1, -1):
            upstream = nodes[i + 1:]
            numedges = random.randint(1, min(maxedges, len(upstream)))
            children = random.sample(upstream, numedges)
            graph[nodes[i]].update(children)
        connected_to_sea = nodes[-1]
        return graph, connected_to_sea

    def sample_instance(self, params: Dict[str, object]) -> Dict[str, object]:
        n = params['num_reservoirs']
        graph, connected_to_sea = self._generate_random_dag(n, params['max_edges'])
        
        reservoirs = [f't{i + 1}' for i in range(n)]
        
        nonfluents = {}
        for t1, out in graph.items():
            for t2 in out:
                nonfluents[f'RES_CONNECT({reservoirs[t1]}, {reservoirs[t2]})'] = True
        nonfluents[f'CONNECTED_TO_SEA({reservoirs[connected_to_sea]})'] = True
        
        states = {}
        for i, res in enumerate(reservoirs):
            top = random.uniform(*params['top_range'])
            range_prop = random.uniform(*params['target_range'])
            range_top = range_prop * top
            lower = random.uniform(0., top - range_top)
            upper = lower + range_top
            nonfluents[f'MIN_LEVEL({res})'] = lower
            nonfluents[f'MAX_LEVEL({res})'] = upper
            nonfluents[f'TOP_RES({res})'] = top
            nonfluents[f'RAIN_VAR({res})'] = params['rain_var']
            states[f'rlevel({res})'] = random.uniform(0., top * 0.9)
        
        return {
            'objects': {'reservoir': reservoirs},
            'non-fluents': nonfluents,
            'init-states': states,
            'horizon': params['horizon'],
            'discount': params['discount'],
            'max-nondef-actions': 'pos-inf'
        }
            
    



import argparse


EPILOG = 'For complete documentation, see https://pyrddlgym.readthedocs.io/en/latest/rddlrepo.html.'

def main():
    parser = argparse.ArgumentParser(prog='rddlrepo',
                                     description="command line parser for the rddlrepository",
                                     epilog=EPILOG)
    subparsers = parser.add_subparsers(dest="rddlrepo", required=True)

    # rebuilding
    parser_build = subparsers.add_parser("build", 
                                         help="(re)build the manifest file that archives all problem paths",
                                         epilog=EPILOG)
    
    # listing
    parser_list = subparsers.add_parser("list",
                                        help="list all problems by context",
                                        epilog=EPILOG)

    # dispatch
    args = parser.parse_args()
    if args.rddlrepo == "build":
        from rddlrepository.core.manager import RDDLRepoManager
        manager = RDDLRepoManager(rebuild=True)
        num_context = len(manager.list_contexts())
        num_problems = len(manager.list_problems())
        print(f'Successfully built rddlrepository manifest: '
              f'found {num_problems} problems across {num_context} contexts.')
        
    elif args.rddlrepo == "list":
        from rddlrepository.core.manager import RDDLRepoManager
        manager = RDDLRepoManager()
        print(manager.get_problems_as_string())
        
    else:
        parser.print_help()

if __name__ == "__main__": 
    main()



info = {
    'name': 'Epidemic',
    'description': 'SIR model with nonlinear infection dynamics, age groups, hospitalization and multiple interventions.',
    'context': '',
    'tags': '',
    'viz': 'None'
}


info = {
    'name': 'MountainCar_Discrete',
    'description': 'A simple continuous MDP with discrete actions for the classical mountain car control problem.',
    'context': 'gym',
    'tags': '',
    'viz': 'MountainCarViz.MountainCarVisualizer'
}


import numpy as np
from PIL import Image
import pygame
from pygame import gfxdraw

from pyRDDLGym.core.compiler.model import RDDLPlanningModel
from pyRDDLGym.core.visualizer.viz import BaseViz


# code comes from openai gym
class AcrobotVisualizer(BaseViz):

    def __init__(self, model: RDDLPlanningModel, screen_dim=500, wait_time=100) -> None:
        self._model = model
        self._wait_time = wait_time
        self.screen_dim = screen_dim
    
    def init_canvas(self, figure_size):
        screen = pygame.Surface(figure_size)
        surf = pygame.Surface(figure_size)
        return screen, surf
        
    def convert2img(self, screen):
        data = np.transpose(np.array(pygame.surfarray.pixels3d(screen)), 
                            axes=(1, 0, 2))
        img = Image.fromarray(data)
        return img
    
    def render(self, state):
        screen, surf = self.init_canvas((self.screen_dim, self.screen_dim))
        surf.fill((255, 255, 255))
        
        bound = 1 + 1 + 0.2 
        scale = self.screen_dim / (bound * 2)
        offset = self.screen_dim // 2
        
        # draw the arm
        p1 = [
            -1 * np.cos(state['theta1']) * scale,
            1 * np.sin(state['theta1']) * scale,
        ]

        p2 = [
            p1[0] - 1 * np.cos(state['theta1'] + state['theta2']) * scale,
            p1[1] + 1 * np.sin(state['theta1'] + state['theta2']) * scale,
        ]

        xys = np.array([[0, 0], p1, p2])[:, ::-1]
        thetas = [state['theta1'] - np.pi / 2, state['theta1'] + state['theta2'] - np.pi / 2]
        link_lengths = [1 * scale, 1 * scale]

        pygame.draw.line(
            surf,
            start_pos=(-2.2 * scale + offset, 1 * scale + offset),
            end_pos=(2.2 * scale + offset, 1 * scale + offset),
            color=(0, 0, 0),
        )
        for ((x, y), th, llen) in zip(xys, thetas, link_lengths):
            x = x + offset
            y = y + offset
            l, r, t, b = 0, llen, 0.1 * scale, -0.1 * scale
            coords = [(l, b), (l, t), (r, t), (r, b)]
            transformed_coords = []
            for coord in coords:
                coord = pygame.math.Vector2(coord).rotate_rad(th)
                coord = (coord[0] + x, coord[1] + y)
                transformed_coords.append(coord)
            gfxdraw.aapolygon(surf, transformed_coords, (0, 204, 204))
            gfxdraw.filled_polygon(surf, transformed_coords, (0, 204, 204))

            gfxdraw.aacircle(surf, int(x), int(y), int(0.1 * scale), (204, 204, 0))
            gfxdraw.filled_circle(surf, int(x), int(y), int(0.1 * scale), (204, 204, 0))
        
        surf = pygame.transform.flip(surf, False, True)
        screen.blit(surf, (0, 0))
        pygame.time.wait(self._wait_time)
        img = self.convert2img(screen)
        del screen, surf        
        
        return img
    



info = {
    'name': 'Tamarisk_POMDP',
    'description': 'The agent manages the spread of an invasive plant species, by manually intervening to eridaticate them or restore the native species. This is the pomdp version',
    'context': 'ippc2014',
    'tags': '',
    'viz': 'None'
}


info = {
    'name': 'Sokoban',
    'description': 'A person pushes boxes in a warehouse onto designated storage areas, difficult domain due to dead ends.',
    'context': 'arcade',
    'tags': '',
    'viz': 'None'
}


info = {
    'name': 'Quadcopter',
    'description': 'Control a swarm of four-propeller drones in 3D space.',
    'context': 'physics',
    'tags': '',
    'viz': 'QuadcopterViz.QuadcopterVisualizer'
}


import matplotlib.pyplot as plt
import matplotlib.patches as patches
from matplotlib.path import Path
import numpy as np
from PIL import Image

from pyRDDLGym.core.compiler.model import RDDLPlanningModel
from pyRDDLGym.core.visualizer.viz import BaseViz


class PongVisualizer(BaseViz):

    def __init__(self, model: RDDLPlanningModel,
                 figure_size=(4, 4),
                 ball_radius=0.02,
                 wait_time=100) -> None:
        self._model = model
        self._figure_size = figure_size
        self._ball_radius = ball_radius
        self._wait_time = wait_time
        
        self._nonfluents = model.ground_vars_with_values(model.non_fluents)
        self._balls = model.type_to_objects['ball']
        
        self.fig = plt.figure(figsize=self._figure_size)
        self.ax = plt.gca()
        
        self.ax.get_xaxis().set_ticks([])
        self.ax.get_yaxis().set_ticks([])
        
    def convert2img(self, fig):
        data = np.frombuffer(fig.canvas.buffer_rgba(), dtype=np.uint8)
        data = data.reshape(fig.canvas.get_width_height()[::-1] + (4,))
        data = data[:, :, :3]
        return Image.fromarray(data)

    def render(self, state):
        
        # draw the balls
        balls = []
        for b in self._balls:
            ball = plt.Circle((state[f'ball-x___{b}'], 
                               state[f'ball-y___{b}']), self._ball_radius, 
                              color='red')
            balls.append(ball)
            self.ax.add_patch(ball)
        
        # draw the paddle
        path = Path([(0.99, state['paddle-y']),
                     (0.99, state['paddle-y'] + self._nonfluents['PADDLE-H'])], 
                    [Path.MOVETO, Path.LINETO])
        paddle = patches.PathPatch(path, color='black', lw=3.0)
        self.ax.add_patch(paddle)
        
        self.fig.canvas.draw()        
        img = self.convert2img(self.fig)
        for ball in balls:
            ball.remove()
        del balls
        paddle.remove()
        del paddle
        
        return img




import math
import numpy as np
from typing import Dict

from pyRDDLGym.Examples.InstanceGenerator import InstanceGenerator


class MountainCarInstanceGenerator(InstanceGenerator):
    
    def get_env_path(self) -> str:
        return 'MountainCar'
    
    def get_domain_name(self) -> str:
        return 'mountain_car'
    
    def sample_instance(self, params: Dict[str, object]) -> Dict[str, object]:
        fx, xrange = self.generate_hills(params['terrain_xleft'], 
                                         params['terrain_widths'],
                                         params['terrain_heights'])
        
        points = self.generate_terrain_points(fx, params['num_points'], *xrange)
        segments = [f's{i + 1}' for i in range(len(points))]
        
        nonfluents = {}
        nonfluents['ACTION-PENALTY'] = 0.0
        nonfluents['MIN-POS'] = xrange[0]
        nonfluents['MAX-POS'] = xrange[1]
        nonfluents['GOAL-MIN'] = params['goal-min']
        for (i, (x1, y1, x2, y2)) in enumerate(points):
            nonfluents[f'X-START({segments[i]})'] = x1
            nonfluents[f'Y-START({segments[i]})'] = y1
            nonfluents[f'X-END({segments[i]})'] = x2
            nonfluents[f'Y-END({segments[i]})'] = y2
        
        states = {}
        states['pos'] = params['pos']
        states['vel'] = params['vel']
        
        return {
            'objects': {'segment': segments},
            'non-fluents': nonfluents,
            'init-states': states,
            'horizon': params['horizon'],
            'discount': params['discount'],
            'max-nondef-actions': 'pos-inf'
        }
    
    def generate_hill(self, x1, x2, h, o1, o2):
        slope = (o2 - o1) / (x2 - x1)
        intercept = o1 - slope * x1
        curve = lambda x: h * (np.sin(slope * x + intercept) + 1.0) / 2.
        return curve
    
    def generate_hills(self, x0, widths, heights):
        x1 = x0
        curves, ranges = [], []
        for i, (w, h) in enumerate(zip(widths, heights)):
            x2 = x1 + w
            if i == 0:
                curve = self.generate_hill(x1, x2, h, o1=-3 * math.pi / 2 * 0.7, o2=-math.pi / 2)
            elif i == len(widths) - 1:
                curve = self.generate_hill(x1, x2, h, o1=-math.pi / 2, o2=math.pi / 2 * 1.2)
            else:
                curve = self.generate_hill(x1, x2, h, o1=-math.pi / 2, o2=3 * math.pi / 2)
            curves.append(curve)
            ranges.append((x1, x2))
            x1 = x2
    
        def piecewise(x):
            for i, (a, b) in enumerate(ranges):
                if x >= a and x < b:
                    return curves[i](x)
            return curves[-1](x)
        
        xrange = (ranges[0][0], ranges[-1][1])
        
        return piecewise, xrange
    
    def generate_terrain_points(self, fx, num_points, xmin, xmax):
        xs = np.linspace(xmin, xmax, num_points)
        ys = np.asarray([fx(x) for x in xs])
        xstart = xs[:-1]
        ystart = ys[:-1]
        xend = xs[1:]
        yend = ys[1:]
        return list(zip(xstart, ystart, xend, yend))
        



info = {
    'name': 'RecSim',
    'description': 'A problem of recommendation systems, with consumers and providers',
    'context': 'ippc2023',
    'tags': '',
    'viz': 'RecSimViz.RecSimVisualizer',
    'generator': 'RecSimInstanceGenerator.RecSimInstanceGenerator'
}


info = {
    'name': 'RaceCar',
    'description': 'A simple continuous MDP for the racecar problem',
    'context': 'ippc2023',
    'tags': '',
    'viz': 'RacecarViz.RacecarVisualizer',
    'generator': 'RaceCarInstanceGenerator.RaceCarInstanceGenerator'
}